# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function

"""The BIM library tool"""

import os
import sys
import tempfile

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")

FILTERS = [
    "*.fcstd",
    "*.FCStd",
    "*.FCSTD",
    "*.stp",
    "*.STP",
    "*.step",
    "*.STEP",
    "*.brp",
    "*.BRP",
    "*.brep",
    "*.BREP",
    "*.ifc",
    "*.IFC",
    "*.sat",
    "*.SAT",
]
TEMPLIBPATH = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "OfflineLibrary")
THUMBNAILSPATH = os.path.join(TEMPLIBPATH, "__thumbcache__")
LIBRARYURL = "https://github.com/FreeCAD/FreeCAD-library/tree/master"
RAWURL = LIBRARYURL.replace("/tree", "/raw")
LIBINDEXFILE = "OfflineLibrary.py"
USE_API = True  # True to use github API instead of web fetching... Way faster
REFRESH_INTERVAL = 3600  # Min seconds between allowing a new API calls (3600 = one hour)


# TODO as https://github.com/yorikvanhavre/BIM_Workbench/pull/77

# All the print() statements in your code should be replaced by
# FreeCAD.Console.PrintMessage() or FreeCAD.Console.PrintWarning() or
# FreeCAD.Console.PrintError() and the text should be placed in a translate()
# function and "\n" should be added to it.
# Example FreeCAD.Console.PrintError(translate("BIM","Please save the document first")+"\n")

# It would be cool if the preview image would have a max width of the available
# column width, so if the task column is smaller than the image, it gets smaller
# to fit the space. I don't remember exactly how to do that, but it should be
# findable in QDesigner


class BIM_Library:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Library",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Library", "Objects Library"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Library", "Opens the objects library"),
        }

    def Activated(self):

        # trying to locate the parts library
        pr = FreeCAD.ParamGet("User parameter:Plugins/parts_library")
        libok = False
        self.librarypath = pr.GetString("destination", "")
        if self.librarypath:
            if os.path.exists(self.librarypath):
                libok = True
        else:
            # check if the library is at the standard addon location
            addondir = os.path.join(FreeCAD.getUserAppDataDir(), "Mod", "parts_library")
            if os.path.exists(addondir):
                # save file paths with forward slashes even on windows
                pr.SetString("destination", addondir.replace("\\", "/"))
                libok = True
        panel = BIM_Library_TaskPanel(offlinemode=libok)
        task = FreeCADGui.Control.showDialog(panel)
        task.setDocumentName(panel.mainDocName)
        task.setAutoCloseOnDeletedDocument(True)


class BIM_Library_TaskPanel:

    def __init__(self, offlinemode=False):

        from PySide import QtGui

        self.mainDocName = FreeCAD.Gui.ActiveDocument.Document.Name
        self.previewDocName = "Viewer"

        self.linked = False

        self.librarypath = FreeCAD.ParamGet("User parameter:Plugins/parts_library").GetString(
            "destination", ""
        )
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogLibrary.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/BIM_Library.svg"))

        # setting up a flat (no directories) file model for search
        self.filemodel = QtGui.QStandardItemModel()
        self.filemodel.setColumnCount(1)

        # setting up a directory model that shows only fcstd, step and brep
        self.dirmodel = LibraryModel()
        self.dirmodel.setRootPath(self.librarypath)
        self.dirmodel.setNameFilters(self.getFilters())
        self.dirmodel.setNameFilterDisables(False)
        self.form.tree.setModel(self.dirmodel)
        self.form.buttonInsert.clicked.connect(self.insert)
        self.form.buttonLink.clicked.connect(self.link)

        self.modelmode = 1  # 0 = File search, 1 = Dir mode

        # Don't show columns for size, file type, and last modified
        self.form.tree.setHeaderHidden(True)
        self.form.tree.hideColumn(1)
        self.form.tree.hideColumn(2)
        self.form.tree.hideColumn(3)
        self.form.tree.setRootIndex(self.dirmodel.index(self.librarypath))
        self.form.searchBox.textChanged.connect(self.onSearch)

        # external search
        sites = {
            "BimObject": [
                "bimobject.png",
                "https://www.bimobject.com/en/product?filetype=8&freetext=",
            ],
            "NBS Library": [
                "nbslibrary.png",
                "https://www.nationalbimlibrary.com/en/search/?facet=Xo-P0w&searchTerm=",
            ],
            "BIMTool": [
                "bimtool.png",
                "https://www.bimtool.com/Catalog.aspx?criterio=",
            ],
            "3DFindIt": ["3dfindit.svg", "https://www.3dfindit.com/textsearch?q="],
            "GrabCAD": [
                "grabcad.svg",
                "https://grabcad.com/library?softwares=step-slash-iges&query=",
            ],
        }
        for k, v in sites.items():
            self.form.comboSearch.addItem(QtGui.QIcon(":/icons/" + v[0]), k, v[1])
        self.form.comboSearch.currentIndexChanged.connect(self.onExternalSearch)

        # retrieve preferences
        self.form.checkOnline.toggled.connect(self.onCheckOnline)
        self.form.checkOnline.setChecked(PARAMS.GetBool("LibraryOnline", not offlinemode))
        self.form.checkFCStdOnly.toggled.connect(self.onCheckFCStdOnly)
        self.form.checkFCStdOnly.setChecked(PARAMS.GetBool("LibraryFCStdOnly", False))
        self.form.check3DPreview.toggled.connect(self.onCheck3DPreview)
        self.form.check3DPreview.setChecked(PARAMS.GetBool("3DPreview", False))

        # collapsables
        if PARAMS.GetBool("LibraryPreview", False):
            self.form.framePreview.show()
            self.form.buttonPreview.setText(translate("BIM", "Preview") + " ▼")
        else:
            self.form.framePreview.hide()
            self.form.buttonPreview.setText(translate("BIM", "Preview") + " ▸")
        self.form.buttonPreview.clicked.connect(self.onButtonPreview)
        self.form.frameOptions.hide()
        self.form.buttonOptions.setText(translate("BIM", "Options") + " ▸")
        self.form.buttonOptions.clicked.connect(self.onButtonOptions)

        # saving functionality, is disabled for now
        self.form.buttonSave.hide()
        self.form.checkThumbnail.hide()
        # self.form.buttonSave.clicked.connect(self.addtolibrary)
        # self.form.checkThumbnail.toggled.connect(self.onCheckThumbnail)
        # self.form.checkThumbnail.setChecked(PARAMS.GetBool("SaveThumbnails",False))
        # self.fcstdCB = QtGui.QCheckBox('FCStd')
        # self.fcstdCB.setCheckState(QtCore.Qt.Checked)
        # self.fcstdCB.setEnabled(False)
        # self.fcstdCB.hide()
        # self.stepCB = QtGui.QCheckBox('STEP')
        # self.stepCB.setCheckState(QtCore.Qt.Checked)
        # self.stepCB.hide()
        # self.stlCB = QtGui.QCheckBox('STL')
        # self.stlCB.setCheckState(QtCore.Qt.Checked)
        # self.stlCB.hide()

        # update the tree
        self.onCheckOnline()

    def onItemSelected(self, selected, deselected):
        """Generates and displays needed previews"""

        from PySide import QtGui

        if not selected:
            return
        index = selected[0].indexes()[0]
        if self.modelmode == 1:
            path = self.dirmodel.filePath(index)
        else:
            path = self.filemodel.itemFromIndex(index).toolTip()
        if path.startswith(":github"):
            path = RAWURL + "/" + path[7:]
        thumb = self.getThumbnail(path)
        if thumb:
            px = QtGui.QPixmap(thumb)
        else:
            px = QtGui.QPixmap()
        self.form.framePreview.setPixmap(px)

        if False:
            # TO BE REFACTORED

            import Part
            import zipfile

            self.previewOn = PARAMS.GetBool("3DPreview", False)
            try:
                self.path = self.dirmodel.filePath(index)
            except:
                self.path = self.previousIndex
                print(self.path)
            self.isFile = os.path.isfile(self.path)
            # if the 3D preview checkbox is on ticked, show the preview
            if self.previewOn == True or self.linked == True:
                if self.isFile == True:
                    # close a non linked preview document
                    if self.linked == False:
                        try:
                            FreeCAD.closeDocument(self.previewDocName)
                        except:
                            pass
                    # create different kinds of previews based on file type
                    if (
                        self.path.lower().endswith(".stp")
                        or self.path.lower().endswith(".step")
                        or self.path.lower().endswith(".brp")
                        or self.path.lower().endswith(".brep")
                    ):
                        self.previewDocName = "Viewer"
                        FreeCAD.newDocument(self.previewDocName)
                        FreeCAD.setActiveDocument(self.previewDocName)
                        Part.show(Part.read(self.path))
                        FreeCADGui.SendMsgToActiveView("ViewFit")
                    elif self.path.lower().endswith(".fcstd"):
                        openedDoc = FreeCAD.openDocument(self.path)
                        FreeCADGui.SendMsgToActiveView("ViewFit")
                        self.previewDocName = FreeCAD.ActiveDocument.Name
                        thumbnailSave = PARAMS.GetBool("SaveThumbnails", False)
                        if thumbnailSave == True:
                            FreeCAD.ActiveDocument.save()
            if self.linked == False:
                self.previousIndex = self.path

            # create a 2D image preview
            if self.path.lower().endswith(".fcstd"):
                zfile = zipfile.ZipFile(self.path)
                files = zfile.namelist()
                # check for meta-file if it's really a FreeCAD document
                if files[0] == "Document.xml":
                    image = "thumbnails/Thumbnail.png"
                    if image in files:
                        image = zfile.read(image)
                        thumbfile = tempfile.mkstemp(suffix=".png")[1]
                        thumb = open(thumbfile, "wb")
                        thumb.write(image)
                        thumb.close()
                        im = QtGui.QPixmap(thumbfile)
                        self.form.framePreview.setPixmap(im)
                        return self.previewDocName, self.previousIndex, self.linked
            self.form.framePreview.clear()
            return self.previewDocName, self.previousIndex, self.linked

    def link(self, index):

        # check if the main document is open
        try:
            # check if the working document is saved
            if FreeCAD.getDocument(self.mainDocName).FileName == "":
                FreeCAD.Console.PrintWarning(
                    translate("BIM", "Save the working file before linking.") + "\n"
                )
            else:
                self.previewOn = PARAMS.GetBool("3DPreview", False)
                self.linked = True
                if self.previewOn != True:
                    BIM_Library_TaskPanel.clicked(self, index, previewDocName="Viewer")
                self.librarypath = ""
                # save the file prior to linking
                BIM_Library_TaskPanel.addtolibrary(self)
                # link a document if it has been previously saved
                if self.fileDialog[0] != "":
                    FreeCADGui.Selection.clearSelection()
                    # link only root objects
                    for obj in FreeCAD.ActiveDocument.RootObjects:
                        FreeCADGui.Selection.addSelection(obj)
                    objects = FreeCADGui.Selection.getSelection()
                    # tries to create a link for each object in the selection
                    for obj in objects:
                        try:
                            link = (
                                FreeCAD.getDocument(self.mainDocName)
                                .addObject("App::Link", "Link")
                                .setLink(obj)
                            )
                            # FreeCAD.getDocument(self.mainDocName).getObject('Link').Label=FreeCAD.ActiveDocument.ActiveObject.Label
                            FreeCAD.getDocument(self.mainDocName).getObject(
                                link
                            ).Label = FreeCAD.ActiveDocument.ActiveObject.Label
                        except:
                            pass
                    FreeCAD.setActiveDocument(self.mainDocName)
                    self.librarypath = FreeCAD.ParamGet(
                        "User parameter:Plugins/parts_library"
                    ).GetString("destination", "")
                    self.linked = False
                    return self.linked
        except:
            FreeCAD.Console.PrintWarning(
                translate("BIM", "It is not possible to link because the main document is closed.")
                + "\n"
            )

    def addtolibrary(self):
        # DISABLED

        import os
        import Mesh
        import Part

        self.fileDialog = QtGui.QFileDialog.getSaveFileName(None, "Save As", self.librarypath)
        # print(self.fileDialog[0])
        # check if file saving has been canceled and save .fcstd, .step and .stl copies
        if self.fileDialog[0] != "":
            # remove the file extension from the file path
            fileName = os.path.splitext(self.fileDialog[0])[0]
            FCfilename = fileName + ".fcstd"
            FreeCAD.ActiveDocument.saveAs(FCfilename)
            if self.stepCB.isChecked() or self.stlCB.isChecked():
                toexport = []
                objs = FreeCAD.ActiveDocument.Objects
                for obj in objs:
                    if obj.ViewObject.Visibility == True:
                        toexport.append(obj)
                if self.stepCB.isChecked() and self.linked == False:
                    STEPfilename = fileName + ".step"
                    Part.export(toexport, STEPfilename)
                if self.stlCB.isChecked() and self.linked == False:
                    STLfilename = fileName + ".stl"
                    Mesh.export(toexport, STLfilename)
        return self.fileDialog[0]

    def onSearch(self, text):

        if text:
            self.setSearchModel(text)
        else:
            self.setFileModel()

    def setSearchModel(self, text):

        from PySide import QtGui

        def add_line(f, dp):
            if self.isAllowed(f) and (text.lower() in f.lower()):
                it = QtGui.QStandardItem(f)
                it.setToolTip(os.path.join(dp, f))
                self.filemodel.appendRow(it)
                if f.lower().endswith(".fcstd"):
                    it.setIcon(QtGui.QIcon(":icons/freecad-doc.png"))
                elif f.lower().endswith(".ifc"):
                    it.setIcon(QtGui.QIcon(":/icons/IFC.svg"))
                else:
                    it.setIcon(QtGui.QIcon(":/icons/Part_document.svg"))

        self.form.tree.setModel(self.filemodel)
        self.filemodel.clear()
        if self.form.checkOnline.isChecked():
            res = self.getOfflineLib(structured=True)
            for i in range(len(res[0])):
                add_line(res[0][i], res[2][i])
        else:
            res = os.walk(self.librarypath)
            for dp, dn, fn in res:
                for f in fn:
                    if not os.path.isdir(os.path.join(dp, f)):
                        add_line(f, dp)
        self.modelmode = 0

    def getFilters(self):

        if self.form.checkFCStdOnly.isChecked():
            return FILTERS
        else:
            return FILTERS[:3]

    def isAllowed(self, filename):

        e = os.path.splitext(filename)[1]
        if e in [f[1:] for f in FILTERS]:
            if e in [f[1:] for f in self.getFilters()]:
                return True
            else:
                return False
        else:
            return True

    def setFileModel(self):

        # self.form.tree.clear()
        self.form.tree.setModel(self.dirmodel)
        self.dirmodel.setRootPath(self.librarypath)
        self.dirmodel.setNameFilters(self.getFilters())
        self.dirmodel.setNameFilterDisables(False)
        self.form.tree.setRootIndex(self.dirmodel.index(self.librarypath))
        self.modelmode = 1
        self.form.tree.setHeaderHidden(True)
        self.form.tree.hideColumn(1)
        self.form.tree.hideColumn(2)
        self.form.tree.hideColumn(3)
        self.form.tree.selectionModel().selectionChanged.connect(self.onItemSelected)

    def setOnlineModel(self):

        from PySide import QtGui

        def addItems(root, d, path):
            for k, v in d.items():
                if self.isAllowed(k):
                    it = QtGui.QStandardItem(k)
                    root.appendRow(it)
                    it.setToolTip(path + "/" + k)
                    if isinstance(v, dict):
                        it.setIcon(
                            QtGui.QIcon.fromTheme("folder", QtGui.QIcon(":/icons/Group.svg"))
                        )
                        addItems(it, v, path + "/" + k)
                        it.setToolTip("")
                    elif k.lower().endswith(".fcstd"):
                        it.setIcon(QtGui.QIcon(":icons/freecad-doc.png"))
                    elif k.lower().endswith(".ifc"):
                        it.setIcon(QtGui.QIcon(":/icons/IFC.svg"))
                    else:
                        it.setIcon(QtGui.QIcon(":/icons/Part_document.svg"))

        self.form.tree.setModel(self.filemodel)
        self.filemodel.clear()
        d = self.getOfflineLib()
        addItems(self.filemodel, d, ":github")
        self.modelmode = 0
        self.form.tree.selectionModel().selectionChanged.connect(self.onItemSelected)

    def getOfflineLib(self, structured=False):

        def addDir(d, root):
            fn = []
            dn = []
            dp = []
            for k, v in d.items():
                if isinstance(v, dict) and v:
                    fn2, dn2, dp2 = addDir(v, root + "/" + k)
                    fn.extend(fn2)
                    dn.extend(dn2)
                    dp.extend(dp2)
                elif v:
                    fn.append(k)
                    dn.append(root)
                    dp.append(root)
            return fn, dn, dp

        templibfile = os.path.join(TEMPLIBPATH, LIBINDEXFILE)
        if not os.path.exists(templibfile):
            FreeCAD.Console.PrintError(
                translate("BIM", "No structure in cache. Refresh required.") + "\n"
            )
            return {}
        import sys

        sys.path.append(TEMPLIBPATH)
        import OfflineLibrary

        d = OfflineLibrary.library
        if structured:
            return addDir(d, ":github")
        else:
            return d

    def urlencode(self, text):

        # print(text, type(text))
        if sys.version_info.major < 3:
            import urllib

            return urllib.quote_plus(text)
        else:
            import urllib.parse

            return urllib.parse.quote_plus(text)

    def openUrl(self, url):

        from PySide import QtGui

        QtGui.QDesktopServices.openUrl(url)

    def needsFullSpace(self):

        return True

    def getStandardButtons(self):

        from PySide import QtGui

        return QtGui.QDialogButtonBox.Close

    def reject(self):

        if hasattr(self, "box") and self.box:
            self.box.off()
        FreeCADGui.Control.closeDialog()
        if self.previewDocName in FreeCAD.listDocuments():
            FreeCAD.closeDocument(self.previewDocName)
        FreeCAD.ActiveDocument.recompute()

    def insert(self, index=None):

        # check if the main document is open
        try:
            FreeCAD.setActiveDocument(self.mainDocName)
        except:
            FreeCAD.Console.PrintError(
                translate(
                    "BIM",
                    "It is not possible to insert this object because the document has been closed.",
                )
                + "\n"
            )
            return
        if not index:
            index = self.form.tree.selectedIndexes()
            if not index:
                return
            index = index[0]
        if self.modelmode == 1:
            path = self.dirmodel.filePath(index)
        else:
            path = self.filemodel.itemFromIndex(index).toolTip()
        if path.startswith(":github"):
            path = self.download(RAWURL + "/" + path[7:])
        before = FreeCAD.ActiveDocument.Objects
        self.name = os.path.splitext(os.path.basename(path))[0]
        ext = os.path.splitext(path.lower())[1]
        if ext in [".stp", ".step", ".brp", ".brep"]:
            self.place(path)
        elif ext == ".fcstd":
            FreeCADGui.ActiveDocument.mergeProject(path)
            from draftutils import todo

            todo.ToDo.delay(self.reject, None)
        elif ext == ".ifc":
            from importers import importIFC

            importIFC.ZOOMOUT = False
            importIFC.insert(path, FreeCAD.ActiveDocument.Name)
            from draftutils import todo

            todo.ToDo.delay(self.reject, None)
        elif ext in [".sat", ".sab"]:
            try:
                # InventorLoader addon
                import importerIL
            except ImportError:
                try:
                    # CADExchanger addon
                    import CadExchangerIO
                except ImportError:
                    FreeCAD.Console.PrintError(
                        translate(
                            "BIM",
                            "Error: Unable to import SAT files - InventorLoader or CadExchanger addon must be installed",
                        )
                        + "\n"
                    )
                else:
                    path = CadExchangerIO.insert(path, FreeCAD.ActiveDocument.Name, returnpath=True)
                    self.place(path)
            else:
                path = importerIL.insert(path, FreeCAD.ActiveDocument.Name)
        FreeCADGui.Selection.clearSelection()
        for o in FreeCAD.ActiveDocument.Objects:
            if not o in before:
                FreeCADGui.Selection.addSelection(o)
        FreeCADGui.SendMsgToActiveView("ViewSelection")

    def download(self, url):

        import urllib.request

        filepath = os.path.join(TEMPLIBPATH, url.split("/")[-1])
        url = url.replace(" ", "%20")
        if not os.path.exists(filepath):
            from PySide import QtCore, QtGui

            QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            u = urllib.request.urlopen(url)
            if not u:
                FreeCAD.Console.PrintError(
                    translate("BIM", "Error: Unable to download") + " " + url + "\n"
                )
            b = u.read()
            f = open(filepath, "wb")
            f.write(b)
            f.close()
            QtGui.QApplication.restoreOverrideCursor()
        return filepath

    def place(self, path):

        import Part
        import WorkingPlane

        self.shape = Part.read(path)
        if hasattr(FreeCADGui, "Snapper"):
            try:
                import DraftTrackers
            except Exception:
                import draftguitools.gui_trackers as DraftTrackers
            self.box = DraftTrackers.ghostTracker(
                self.shape, dotted=True, scolor=(0.0, 0.0, 1.0), swidth=1.0
            )
            self.delta = self.shape.BoundBox.Center
            self.box.move(self.delta)
            self.box.on()
            WorkingPlane.get_working_plane()
            self.origin = self.makeOriginWidget()
            FreeCADGui.Snapper.getPoint(
                movecallback=self.mouseMove,
                callback=self.mouseClick,
                extradlg=self.origin,
            )
        else:
            Part.show(self.shape)

    def makeOriginWidget(self):

        from PySide import QtGui

        w = QtGui.QWidget()
        w.setWindowTitle(translate("BIM", "Insertion point"))
        w.setWindowIcon(
            QtGui.QIcon(os.path.join(os.path.dirname(__file__), "icons", "BIM_Library.svg"))
        )
        l = QtGui.QVBoxLayout()
        w.setLayout(l)
        c = QtGui.QComboBox()
        c.ObjectName = "comboOrigin"
        w.comboOrigin = c
        c.addItems(
            [
                translate("BIM", "Origin"),
                translate("BIM", "Top left"),
                translate("BIM", "Top center"),
                translate("BIM", "Top right"),
                translate("BIM", "Middle left"),
                translate("BIM", "Middle center"),
                translate("BIM", "Middle right"),
                translate("BIM", "Bottom left"),
                translate("BIM", "Bottom center"),
                translate("BIM", "Bottom right"),
            ]
        )
        c.setCurrentIndex(PARAMS.GetInt("LibraryDefaultInsert", 0))
        c.currentIndexChanged.connect(self.storeInsert)
        l.addWidget(c)
        return w

    def storeInsert(self, index):

        PARAMS.SetInt("LibraryDefaultInsert", index)

    def mouseMove(self, point, info):

        self.box.move(point.add(self.getDelta()))

    def mouseClick(self, point, info):

        if point:
            import Arch

            self.box.off()
            self.shape.translate(point.add(self.getDelta()))
            obj = Arch.makeEquipment()
            obj.Shape = self.shape
            obj.Label = self.name
        self.reject()

    def getDelta(self):

        d = FreeCAD.Vector(-self.shape.BoundBox.Center.x, -self.shape.BoundBox.Center.y, 0)
        idx = self.origin.comboOrigin.currentIndex()
        if idx <= 0:
            return FreeCAD.Vector()
        elif idx == 1:
            return d.add(
                FreeCAD.Vector(self.shape.BoundBox.XLength / 2, -self.shape.BoundBox.YLength / 2, 0)
            )
        elif idx == 2:
            return d.add(FreeCAD.Vector(0, -self.shape.BoundBox.YLength / 2, 0))
        elif idx == 3:
            return d.add(
                FreeCAD.Vector(
                    -self.shape.BoundBox.XLength / 2,
                    -self.shape.BoundBox.YLength / 2,
                    0,
                )
            )
        elif idx == 4:
            return d.add(FreeCAD.Vector(self.shape.BoundBox.XLength / 2, 0, 0))
        elif idx == 5:
            return d
        elif idx == 6:
            return d.add(FreeCAD.Vector(-self.shape.BoundBox.XLength / 2, 0, 0))
        elif idx == 7:
            return d.add(
                FreeCAD.Vector(self.shape.BoundBox.XLength / 2, self.shape.BoundBox.YLength / 2, 0)
            )
        elif idx == 8:
            return d.add(FreeCAD.Vector(0, self.shape.BoundBox.YLength / 2, 0))
        elif idx == 9:
            return d.add(
                FreeCAD.Vector(-self.shape.BoundBox.XLength / 2, self.shape.BoundBox.YLength / 2, 0)
            )

    def getOnlineContentsAPI(self, url):
        """same as getOnlineContents but uses github API (faster)"""

        import json
        import requests

        result = {}
        count = 0
        r = requests.get(
            "https://api.github.com/repos/FreeCAD/FreeCAD-library/git/trees/master?recursive=1"
        )
        if r.ok:
            j = json.loads(r.content)
            if j["truncated"]:
                print(
                    "WARNING: The fetched content exceeds maximum GitHub allowance and is truncated"
                )
            t = j["tree"]
            for f in t:
                path = f["path"].split("/")
                if f["type"] == "tree":
                    name = None
                else:
                    name = path[-1]
                    path = path[:-1]
                host = result
                for fp in path:
                    if fp in host:
                        host = host[fp]
                    else:
                        host[fp] = {}
                        host = host[fp]
                if name:
                    for ft in self.getFilters():
                        if name.endswith(ft[1:]):
                            break
                    else:
                        continue
                    host[name] = name
                    count += 1
        else:
            FreeCAD.Console.PrintError(translate("BIM", "Could not fetch library contents") + "\n")
        # print("result:",result)
        if not result:
            FreeCAD.Console.PrintError(
                translate("BIM", "No results fetched from online library") + "\n"
            )
        else:
            FreeCAD.Console.PrintLog("BIM Library: Reloaded " + str(count) + " files\n")
        return result

    def onCheckOnline(self, state=None):
        """if the Online checkbox is clicked"""

        import datetime

        if state == None:
            state = self.form.checkOnline.isChecked()
        # save state
        PARAMS.SetBool("LibraryOnline", state)
        if state:
            # online
            if USE_API:
                needrefresh = True
                timestamp = datetime.datetime.now()
                if os.path.exists(os.path.join(TEMPLIBPATH, LIBINDEXFILE)):
                    stored = PARAMS.GetUnsigned("LibraryTimeStamp", 0)
                    if stored:
                        stored = datetime.datetime.fromtimestamp(stored)
                        if (timestamp - stored).total_seconds() < REFRESH_INTERVAL:
                            needrefresh = False
                if needrefresh:
                    PARAMS.SetUnsigned("LibraryTimeStamp", int(timestamp.timestamp()))
                    self.onRefresh()
                else:
                    FreeCAD.Console.PrintLog("BIM Library: Using cached library\n")
            self.setOnlineModel()
            self.form.buttonLink.setEnabled(False)
        else:
            # offline
            self.setFileModel()
            self.form.buttonLink.setEnabled(True)

    def onRefresh(self):
        """refreshes the tree"""

        from PySide import QtCore, QtGui

        def writeOfflineLib():
            if USE_API:
                rootfiles = self.getOnlineContentsAPI(LIBRARYURL)
            if rootfiles:
                templibfile = os.path.join(TEMPLIBPATH, LIBINDEXFILE)
                os.makedirs(TEMPLIBPATH, exist_ok=True)
                tf = open(templibfile, "w", encoding="utf8")
                tf.write("library=" + str(rootfiles) + "\n")
                tf.close()
                self.setOnlineModel()

        reply = PARAMS.GetBool("LibraryWarning", False)
        if not reply:
            reply = QtGui.QMessageBox.information(
                None, "", translate("BIM", "Warning, this can take several minutes!")
            )
        if reply:
            PARAMS.SetBool("LibraryWarning", True)
            self.form.setEnabled(False)
            QtGui.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
            self.form.repaint()
            QtGui.QApplication.processEvents()
            QtCore.QTimer.singleShot(1, writeOfflineLib)
            self.form.setEnabled(True)
            QtGui.QApplication.restoreOverrideCursor()
        else:
            self.setOnlineModel()

    def onCheckFCStdOnly(self, state):
        """if the FCStd only checkbox is clicked"""

        # save state
        PARAMS.SetBool("LibraryFCStdOnly", state)
        self.dirmodel.setNameFilters(self.getFilters())
        self.onCheckOnline(self.form.checkOnline.isChecked())

    def onCheck3DPreview(self, state):
        """if the 3D preview checkbox is clicked"""

        # save state
        PARAMS.SetBool("3DPreview", state)
        self.previewOn = PARAMS.GetBool("3DPreview", False)
        try:
            FreeCAD.closeDocument(self.previewDocName)
        except:
            pass
        if self.previewOn == True:
            self.previewDocName = "Viewer"
            self.doc = FreeCAD.newDocument(self.previewDocName)
            FreeCADGui.ActiveDocument.ActiveView.viewIsometric()
            return self.previewDocName

    def onCheckThumbnail(self, state):
        """if the thumbnail checkbox is clicked"""

        # save state
        PARAMS.SetBool("SaveThumbnails", state)

    def onButtonOptions(self):
        """hides/shows the options"""

        if self.form.frameOptions.isVisible():
            self.form.frameOptions.hide()
            self.form.buttonOptions.setText(translate("BIM", "Options") + " ▸")
        else:
            self.form.frameOptions.show()
            self.form.buttonOptions.setText(translate("BIM", "Options") + " ▼")

    def onButtonPreview(self):
        """hides/shows the preview"""

        if self.form.framePreview.isVisible():
            self.form.framePreview.hide()
            self.form.buttonPreview.setText(translate("BIM", "Preview") + " ▸")
            PARAMS.SetBool("LibraryPreview", False)
        else:
            self.form.framePreview.show()
            self.form.buttonPreview.setText(translate("BIM", "Preview") + " ▼")
            PARAMS.SetBool("LibraryPreview", True)

    def getThumbnail(self, filepath):
        """returns a thumbnail image path for a given file path"""

        import urllib.request
        import urllib.parse
        import zipfile
        import io

        if not filepath.lower().endswith(".fcstd"):
            return None
        iconname = self.getHashname(filepath)
        iconfile = os.path.join(THUMBNAILSPATH, iconname)
        if os.path.exists(iconfile):
            return iconfile
        else:
            if self.form.checkOnline.isChecked():
                # download file
                u = urllib.request.urlopen(urllib.parse.quote(filepath, safe=":/."))
                fdata = u.read()
                u.close()
                f = io.BytesIO(fdata)
            else:
                f = filepath
            zfile = zipfile.ZipFile(f)
            if "thumbnails/Thumbnail.png" in zfile.namelist():
                data = zfile.read("thumbnails/Thumbnail.png")
                os.makedirs(os.path.dirname(iconfile), exist_ok=True)
                thumb = open(iconfile, "wb")
                thumb.write(data)
                thumb.close()
                return iconfile
            else:
                return None

    def getHashname(self, filepath):
        """creates a png filename for a given file path"""

        import hashlib

        filepath = self.cleanPath(filepath)
        return hashlib.md5(filepath.encode()).hexdigest() + ".png"

    def cleanPath(self, filepath):
        """cleans a file path into subfolder/subfolder/file form"""

        import urllib.request
        import urllib.parse

        if filepath.startswith(self.librarypath):
            # strip local part od the path
            filepath = filepath[len(self.librarypath) :]
        if filepath.startswith(RAWURL):
            filepath = filepath[len(RAWURL) :]
        filepath = filepath.replace("\\", "/")
        if filepath.startswith("/"):
            filepath = filepath[1:]
        filepath = urllib.parse.quote(filepath)
        return filepath

    def onExternalSearch(self, index):
        """searches on external websites"""

        if index > 0:
            baseurl = self.form.comboSearch.itemData(index)
            term = self.form.searchBox.text()
            if term:
                self.openUrl(baseurl + self.urlencode(term))


if FreeCAD.GuiUp:

    from PySide import QtCore, QtGui

    class LibraryModel(QtGui.QFileSystemModel):
        "a custom QFileSystemModel that displays FreeCAD file icons"

        def __init__(self):

            QtGui.QFileSystemModel.__init__(self)

        def data(self, index, role):

            if index.column() == 0 and role == QtCore.Qt.DecorationRole:
                if index.data().lower().endswith(".fcstd"):
                    return QtGui.QIcon(":icons/freecad-doc.png")
                elif index.data().lower().endswith(".ifc"):
                    return QtGui.QIcon(os.path.join(os.path.dirname(__file__), "icons", "IFC.svg"))
                elif index.data().lower() == "private":
                    return QtGui.QIcon.fromTheme("folder-lock")
            return super(LibraryModel, self).data(index, role)


FreeCADGui.addCommand("BIM_Library", BIM_Library())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2023 FreeCAD Project Association                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to set up the working plane and its grid."""
## @package gui_selectplane
# \ingroup draftguitools
# \brief Provides GUI tools to set up the working plane and its grid.

## \addtogroup draftguitools
# @{
from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Part
import WorkingPlane

from FreeCAD import Units
from drafttaskpanels import task_selectplane
from draftutils import gui_utils
from draftutils import params
from draftutils import utils
from draftutils.messages import _toolmsg
from draftutils.todo import todo
from draftutils.translate import translate

__title__ = "FreeCAD Draft Workbench GUI Tools - Working plane-related tools"
__author__ = "Yorik van Havre, Werner Mayer, Martin Burbaum, Ken Cline, " "Dmitry Chigrin"
__url__ = "https://www.freecad.org"


class Draft_SelectPlane:
    """The Draft_SelectPlane FreeCAD command definition."""

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_SelectPlane",
            "Accel": "W, P",
            "MenuText": QT_TRANSLATE_NOOP("Draft_SelectPlane", "Working Plane"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_SelectPlane",
                "Defines the working plane from 3 vertices, 1 or more shapes, or an object",
            ),
        }

    def IsActive(self):
        """Return True when this command should be available."""
        return bool(gui_utils.get_3d_view())

    def Activated(self):
        """Execute when the command is called."""
        # Finish active Draft command if any
        if App.activeDraftCommand is not None:
            App.activeDraftCommand.finish()

        App.activeDraftCommand = self
        self.call = None

        # Set variables
        self.wp = WorkingPlane.get_working_plane()
        self.view = self.wp._view
        self.grid = None
        if hasattr(Gui, "Snapper"):
            Gui.Snapper.setTrackers()
            self.grid = Gui.Snapper.grid
        self.offset = 0
        self.center = params.get_param("CenterPlaneOnView")

        # Create task panel
        self.taskd = task_selectplane.SelectPlaneTaskPanel()
        self.taskd.reject = self.reject
        form = self.taskd.form

        # Fill values
        form.fieldOffset.setText(Units.Quantity(self.offset, Units.Length).UserString)
        form.checkCenter.setChecked(self.center)
        try:
            q = Units.Quantity(params.get_param("gridSpacing"))
        except ValueError:
            q = Units.Quantity("1 mm")
        form.fieldGridSpacing.setText(q.UserString)
        form.fieldGridMainLine.setValue(params.get_param("gridEvery"))
        form.fieldGridExtension.setValue(params.get_param("gridSize"))
        form.fieldSnapRadius.setValue(params.get_param("snapRange"))

        # Set icons
        form.setWindowIcon(QtGui.QIcon(":/icons/Draft_SelectPlane.svg"))
        form.buttonTop.setIcon(QtGui.QIcon(":/icons/view-top.svg"))
        form.buttonFront.setIcon(QtGui.QIcon(":/icons/view-front.svg"))
        form.buttonSide.setIcon(QtGui.QIcon(":/icons/view-right.svg"))
        form.buttonAlign.setIcon(QtGui.QIcon(":/icons/view-isometric.svg"))
        form.buttonAuto.setIcon(QtGui.QIcon(":/icons/view-axonometric.svg"))
        form.buttonMove.setIcon(QtGui.QIcon(":/icons/Draft_Move.svg"))
        form.buttonCenter.setIcon(QtGui.QIcon(":/icons/view-fullscreen.svg"))
        form.buttonPrevious.setIcon(QtGui.QIcon(":/icons/sel-back.svg"))
        form.buttonNext.setIcon(QtGui.QIcon(":/icons/sel-forward.svg"))

        # Grid color
        color = params.get_param("gridColor")
        form.buttonColor.setProperty("color", QtGui.QColor(utils.rgba_to_argb(color)))

        # Connect slots
        form.buttonTop.clicked.connect(self.on_click_top)
        form.buttonFront.clicked.connect(self.on_click_front)
        form.buttonSide.clicked.connect(self.on_click_side)
        form.buttonAlign.clicked.connect(self.on_click_align)
        form.buttonAuto.clicked.connect(self.on_click_auto)
        form.buttonMove.clicked.connect(self.on_click_move)
        form.buttonCenter.clicked.connect(self.on_click_center)
        form.buttonPrevious.clicked.connect(self.on_click_previous)
        form.buttonNext.clicked.connect(self.on_click_next)
        form.fieldOffset.textEdited.connect(self.on_set_offset)
        if hasattr(form.checkCenter, "checkStateChanged"):  # Qt version >= 6.7.0
            form.checkCenter.checkStateChanged.connect(self.on_set_center)
        else:  # Qt version < 6.7.0
            form.checkCenter.stateChanged.connect(self.on_set_center)
        form.fieldGridSpacing.textEdited.connect(self.on_set_grid_size)
        form.fieldGridMainLine.valueChanged.connect(self.on_set_main_line)
        form.fieldGridExtension.valueChanged.connect(self.on_set_extension)
        form.fieldSnapRadius.valueChanged.connect(self.on_set_snap_radius)
        form.buttonColor.changed.connect(self.on_color_changed)

        # Enable/disable buttons.
        form.buttonPrevious.setEnabled(self.wp._has_previous())
        form.buttonNext.setEnabled(self.wp._has_next())

        # Try to find a WP from the current selection
        if Gui.Selection.hasSelection():
            if self.wp.align_to_selection(self.offset):
                Gui.Selection.clearSelection()
            self.finish()
            return

        # Execute the actual task panel delayed to catch possible active Draft command
        todo.delay(Gui.Control.showDialog, self.taskd)
        todo.delay(form.setFocus, None)
        _toolmsg(
            translate(
                "draft",
                "Select 3 vertices, one or more shapes or an object to define a working plane",
            )
        )
        self.call = self.view.addEventCallback("SoEvent", self.action)

    def finish(self):
        """Execute when the command is terminated."""
        App.activeDraftCommand = None
        Gui.Control.closeDialog()
        if hasattr(Gui, "Snapper"):
            Gui.Snapper.off()
        # Terminate coin callbacks
        if self.call:
            try:
                self.view.removeEventCallback("SoEvent", self.call)
            except RuntimeError:
                # The view has been deleted already
                pass
            self.call = None

    def reject(self):
        """Execute when clicking the Cancel button."""
        self.finish()

    def action(self, arg):
        """Set the callbacks for the view."""
        if arg["Type"] == "SoKeyboardEvent" and arg["Key"] == "ESCAPE":
            self.reject()
        if (
            arg["Type"] == "SoMouseButtonEvent"
            and (arg["State"] == "UP")
            and (arg["Button"] == "BUTTON1")
        ):
            self.check_selection()

    def check_selection(self):
        """Check the selection, if it is usable, finish the command."""
        if self.wp.align_to_selection(self.offset):
            Gui.Selection.clearSelection()
            self.finish()

    def on_click_top(self):
        self.wp.set_to_top(self.offset, self.center)
        self.finish()

    def on_click_front(self):
        self.wp.set_to_front(self.offset, self.center)
        self.finish()

    def on_click_side(self):
        self.wp.set_to_side(self.offset, self.center)
        self.finish()

    def on_click_align(self):
        self.wp.set_to_view(self.offset, self.center)
        self.finish()

    def on_click_auto(self):
        self.wp.set_to_auto()
        self.finish()

    def on_click_move(self):
        sels = Gui.Selection.getSelectionEx("", 0)
        if (
            len(sels) == 1
            and len(sels[0].SubObjects) == 1
            and sels[0].SubObjects[0].ShapeType == "Vertex"
        ):
            vert = Part.getShape(
                sels[0].Object, sels[0].SubElementNames[0], needSubElement=True, retType=0
            )
            self.wp.set_to_position(vert.Point)
            Gui.Selection.clearSelection()
            self.finish()
        else:
            # Move the WP to the center of the current view
            self.wp.center_on_view()
            self.finish()

    def on_click_center(self):
        self.wp.align_view()
        self.finish()

    def on_click_previous(self):
        self.wp._previous()
        self.finish()

    def on_click_next(self):
        self.wp._next()
        self.finish()

    def on_set_offset(self, text):
        try:
            q = Units.Quantity(text)
        except Exception:
            pass
        else:
            self.offset = q.Value

    def on_set_center(self, val):
        self.center = bool(getattr(val, "value", val))
        params.set_param("CenterPlaneOnView", self.center)

    def on_set_grid_size(self, text):
        try:
            q = Units.Quantity(text)
        except Exception:
            pass
        else:
            params.set_param("gridSpacing", q.UserString)
            # ParamObserver handles grid changes. See params.py.
            if self.grid is not None:
                self.grid.show_during_command = True
                self.grid.on()

    def on_set_main_line(self, i):
        if i > 1:
            params.set_param("gridEvery", i)
            # ParamObserver handles grid changes. See params.py.
            if self.grid is not None:
                self.grid.show_during_command = True
                self.grid.on()

    def on_set_extension(self, i):
        if i > 1:
            params.set_param("gridSize", i)
            # ParamObserver handles grid changes. See params.py.
            if self.grid is not None:
                self.grid.show_during_command = True
                self.grid.on()

    def on_set_snap_radius(self, i):
        params.set_param("snapRange", i)
        if hasattr(Gui, "Snapper"):
            Gui.Snapper.showradius()

    def on_color_changed(self):
        color = utils.argb_to_rgba(self.taskd.form.buttonColor.property("color").rgba())
        params.set_param("gridColor", color)


Gui.addCommand("Draft_SelectPlane", Draft_SelectPlane())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import PathScripts.PathUtils as PathUtils
import unittest


class depthTestCases(unittest.TestCase):
    def test00(self):
        """Stepping down to zero"""
        args = {
            "clearance_height": 15,
            "safe_height": 12,
            "start_depth": 10,
            "step_down": 2,
            "z_finish_step": 1,
            "final_depth": 0,
            "user_depths": None,
        }

        expected = [8, 6, 4, 2, 1, 0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test001(self):
        """Stepping from zero to a negative depth"""

        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 0,
            "step_down": 2,
            "z_finish_step": 0,
            "final_depth": -10,
            "user_depths": None,
        }

        expected = [-2, -4, -6, -8, -10]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test002(self):
        """Start and end are equal or start lower than finish"""

        args = {
            "clearance_height": 15,
            "safe_height": 12,
            "start_depth": 10,
            "step_down": 2,
            "z_finish_step": 0,
            "final_depth": 10,
            "user_depths": None,
        }
        expected = [10]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

        args["start_depth"] = 10
        args["final_depth"] = 15

        expected = []

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test003(self):
        """User Parameters passed in"""
        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 0,
            "step_down": 2,
            "z_finish_step": 0,
            "final_depth": -10,
            "user_depths": [2, 4, 8, 10, 11, 12],
        }

        expected = [2, 4, 8, 10, 11, 12]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test004(self):
        """z_finish_step passed in."""
        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 0,
            "step_down": 2,
            "z_finish_step": 1,
            "final_depth": -10,
            "user_depths": None,
        }

        expected = [-2, -4, -6, -8, -9, -10]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test005(self):
        """stepping down with equalstep=True"""
        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 10,
            "step_down": 3,
            "z_finish_step": 0,
            "final_depth": 0,
            "user_depths": None,
            "equalstep": True,
        }

        expected = [7.5, 5.0, 2.5, 0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test006(self):
        """stepping down with equalstep=True and a finish depth"""
        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 10,
            "step_down": 3,
            "z_finish_step": 1,
            "final_depth": 0,
            "user_depths": None,
        }

        expected = [7.0, 4.0, 1.0, 0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test007(self):
        """stepping down with stepdown greater than total depth"""
        args = {
            "clearance_height": 10,
            "safe_height": 5,
            "start_depth": 10,
            "step_down": 20,
            "z_finish_step": 1,
            "final_depth": 0,
            "user_depths": None,
        }

        expected = [1.0, 0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

    def test008(self):
        """Test handling of negative step-down, negative finish step, and relative size of step/finish"""

        # negative steps should be converted to positive values
        args = {
            "clearance_height": 3,
            "safe_height": 3,
            "start_depth": 2,
            "step_down": -1,
            "z_finish_step": -1,
            "final_depth": 0,
            "user_depths": None,
        }

        expected = [1.0, 0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected)

        # a step_down less than the finish step is an error
        args = {
            "clearance_height": 3,
            "safe_height": 3,
            "start_depth": 2,
            "step_down": 0.1,
            "z_finish_step": 1,
            "final_depth": 0,
            "user_depths": None,
        }
        self.assertRaises(ValueError, PathUtils.depth_params, **args)

    def test009(self):
        """stepping down with single stepdown exactly equal to total depth"""
        args = {
            "clearance_height": 20.0,
            "safe_height": 15.0,
            "start_depth": 10.0,
            "step_down": 10.0,
            "z_finish_step": 0.0,
            "final_depth": 0.0,
            "user_depths": None,
        }

        expected = [0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected, "Expected {}, but result of {}".format(expected, r))

    def test010(self):
        """stepping down with single stepdown roughly equal to total depth"""
        args = {
            "clearance_height": 20.0,
            "safe_height": 15.0,
            "start_depth": 10.000000001,
            "step_down": 10.0,
            "z_finish_step": 0.0,
            "final_depth": 0.0,
            "user_depths": None,
        }

        expected = [0]

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected, "Expected {}, but result of {}".format(expected, r))

        args = {
            "clearance_height": 20.0,
            "safe_height": 15.0,
            "start_depth": 10.0,
            "step_down": 9.9999999,
            "z_finish_step": 0.0,
            "final_depth": 0.0,
            "user_depths": None,
        }

        d = PathUtils.depth_params(**args)
        r = [i for i in d]
        self.assertListEqual(r, expected, "Expected {}, but result of {}".format(expected, r))



# SPDX-License-Identifier: LGPL-2.1-or-later
# /**************************************************************************
#                                                                           *
#    Copyright (c) 2023 Ondsel <development@ondsel.com>                     *
#                                                                           *
#    This file is part of FreeCAD.                                          *
#                                                                           *
#    FreeCAD is free software: you can redistribute it and/or modify it     *
#    under the terms of the GNU Lesser General Public License as            *
#    published by the Free Software Foundation, either version 2.1 of the   *
#    License, or (at your option) any later version.                        *
#                                                                           *
#    FreeCAD is distributed in the hope that it will be useful, but         *
#    WITHOUT ANY WARRANTY; without even the implied warranty of             *
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
#    Lesser General Public License for more details.                        *
#                                                                           *
#    You should have received a copy of the GNU Lesser General Public       *
#    License along with FreeCAD. If not, see                                *
#    <https://www.gnu.org/licenses/>.                                       *
#                                                                           *
# **************************************************************************/

import re
import os
import FreeCAD as App

from PySide.QtCore import QT_TRANSLATE_NOOP

if App.GuiUp:
    import FreeCADGui as Gui
    from PySide import QtCore, QtGui, QtWidgets
    from PySide.QtWidgets import QPushButton, QMenu

import UtilsAssembly
import Preferences
from functools import partial

__title__ = "Assembly Command Create Bill of Materials"
__author__ = "Ondsel"
__url__ = "https://www.freecad.org"

translate = App.Qt.translate

TranslatedColumnNames = [
    translate("Assembly", "Index (auto)"),
    translate("Assembly", "Name (auto)"),
    translate("Assembly", "Description"),
    translate("Assembly", "File Name (auto)"),
    translate("Assembly", "Quantity (auto)"),
]

ColumnNames = [
    "Index",
    "Name",
    "Description",
    "File Name",
    "Quantity",
]


class CommandCreateBom:
    def __init__(self):
        pass

    def GetResources(self):
        return {
            "Pixmap": "Assembly_BillOfMaterials",
            "MenuText": QT_TRANSLATE_NOOP("Assembly_CreateBom", "Bill of Materials"),
            "Accel": "O",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Assembly_CreateBom",
                "<p>Creates a bill of materials of the current assembly. If an assembly is active, it will be a BOM of this assembly. Else it will be a BOM of the whole document.</p>"
                "<p>The BOM object is a document object that stores the settings of your BOM. It is also a spreadsheet object so you can easily visualize the BOM. If you do not need the BOM object to be saved as a document object, you can simply export and cancel the task.</p>"
                "<p>The columns 'Index', 'Name', 'File Name' and 'Quantity' are automatically generated on recompute. The 'Description' and custom columns are not overwritten.</p>",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        return True

    def Activated(self):
        self.panel = TaskAssemblyCreateBom()
        dialog = Gui.Control.showDialog(self.panel)
        if dialog is not None:
            dialog.setAutoCloseOnDeletedDocument(True)
            dialog.setDocumentName(App.ActiveDocument.Name)


######### Create Exploded View Task ###########
class TaskAssemblyCreateBom(QtCore.QObject):
    def __init__(self, bomObj=None):
        super().__init__()

        self.form = Gui.PySideUic.loadUi(":/panels/TaskAssemblyCreateBom.ui")

        # Set the QListWidget properties to support drag and drop
        self.form.columnList.setEditTriggers(
            QtWidgets.QAbstractItemView.DoubleClicked | QtWidgets.QAbstractItemView.EditKeyPressed
        )
        self.form.columnList.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.form.columnList.setDragEnabled(True)
        self.form.columnList.setAcceptDrops(True)
        self.form.columnList.setDropIndicatorShown(True)
        self.form.columnList.setDragDropMode(QtWidgets.QAbstractItemView.InternalMove)

        self.form.columnList.installEventFilter(self)

        self.form.btnAddColumn.clicked.connect(self.showAddColumnMenu)
        self.form.btnExport.clicked.connect(self.export)

        self.form.helpButton.clicked.connect(self.showHelpDialog)

        pref = Preferences.preferences()

        if bomObj:
            App.setActiveTransaction("Edit Bill Of Materials")

            for name in bomObj.columnsNames:
                if name in ColumnNames:
                    index = ColumnNames.index(name)
                    name = TranslatedColumnNames[index]

                self.addColItem(name)

            self.bomObj = bomObj
        else:
            App.setActiveTransaction("Create Bill Of Materials")

            # Add the columns
            for name in TranslatedColumnNames:
                self.addColItem(name)

            self.createBomObject()
            self.bomObj.onlyParts = pref.GetBool("BOMOnlyParts", False)
            self.bomObj.detailParts = pref.GetBool("BOMDetailParts", True)
            self.bomObj.detailSubAssemblies = pref.GetBool("BOMDetailSubAssemblies", True)

        self.form.CheckBox_onlyParts.setChecked(self.bomObj.onlyParts)
        self.form.CheckBox_detailParts.setChecked(self.bomObj.detailParts)
        self.form.CheckBox_detailSubAssemblies.setChecked(self.bomObj.detailSubAssemblies)

        self.form.columnList.model().rowsMoved.connect(self.onItemsReordered)
        self.form.columnList.itemChanged.connect(self.itemUpdated)

        self.form.CheckBox_onlyParts.stateChanged.connect(self.onIncludeSolids)
        self.form.CheckBox_detailParts.stateChanged.connect(self.onDetailParts)
        self.form.CheckBox_detailSubAssemblies.stateChanged.connect(self.onDetailSubAssemblies)

        self.updateColumnList()

    def accept(self):
        self.deactivate()
        App.closeActiveTransaction()

        self.bomObj.recompute()

        self.bomObj.ViewObject.showSheetMdi()

        return True

    def reject(self):
        self.deactivate()
        App.closeActiveTransaction(True)
        return True

    def deactivate(self):
        pref = Preferences.preferences()
        pref.SetBool("BOMOnlyParts", self.form.CheckBox_onlyParts.isChecked())
        pref.SetBool("BOMDetailParts", self.form.CheckBox_detailParts.isChecked())
        pref.SetBool("BOMDetailSubAssemblies", self.form.CheckBox_detailSubAssemblies.isChecked())

        if Gui.Control.activeDialog():
            Gui.Control.closeDialog()

    def onIncludeSolids(self, val):
        self.bomObj.onlyParts = val

    def onDetailParts(self, val):
        self.bomObj.detailParts = val

    def onDetailSubAssemblies(self, val):
        self.bomObj.detailSubAssemblies = val

    def addColumn(self):
        new_name = translate("Assembly", "Default")
        if self.isNameDuplicate(new_name):
            # Find a unique name
            counter = 1
            while self.isNameDuplicate(f"{new_name}_{counter}"):
                counter += 1
            new_name = f"{new_name}_{counter}"

        item = self.addColItem(new_name)

        # Ensure the new item is selected and starts editing
        self.form.columnList.setCurrentItem(item)
        self.form.columnList.editItem(item)
        self.updateColumnList()

    def addColItem(self, name):
        item = QtWidgets.QListWidgetItem(name)

        isCustomCol = self.isCustomColumn(name)

        if isCustomCol:
            item.setFlags(item.flags() | QtCore.Qt.ItemIsEditable)
        else:
            font = item.font()
            font.setBold(True)
            item.setFont(font)

        self.form.columnList.addItem(item)
        return item

    def showAddColumnMenu(self):
        menu = QtWidgets.QMenu()
        # Get the current columns in the list
        current_columns = [
            self.form.columnList.item(i).text() for i in range(self.form.columnList.count())
        ]

        # Add actions for columns that are not currently in the list
        noneAdded = True
        for name in TranslatedColumnNames:
            if name not in current_columns:
                action = QtGui.QAction(f"Add '{name}' column", self)
                action.triggered.connect(partial(self.addColItem, name))
                menu.addAction(action)
                noneAdded = False

        if noneAdded:
            self.addColumn()
            return

        # Add the action for adding a custom column
        action = QtGui.QAction("Add custom column", self)
        action.triggered.connect(self.addColumn)
        menu.addAction(action)

        # Show the menu below the button
        menu.exec_(
            self.form.btnAddColumn.mapToGlobal(QtCore.QPoint(0, self.form.btnAddColumn.height()))
        )

    def isCustomColumn(self, name):
        isCustomCol = True
        if name in TranslatedColumnNames:
            # Description column is currently not auto generated so it's a custom column
            index = TranslatedColumnNames.index(name)
            if ColumnNames[index] != "Description":
                isCustomCol = False
        return isCustomCol

    def onItemsReordered(self, parent, start, end, destination, row):
        self.updateColumnList()

    def updateColumnList(self):
        if self.bomObj:
            new_names = []
            for i in range(self.form.columnList.count()):
                text = self.form.columnList.item(i).text()
                if text in TranslatedColumnNames:
                    index = TranslatedColumnNames.index(text)
                    text = ColumnNames[index]
                new_names.append(text)
            self.bomObj.columnsNames = new_names

    def itemUpdated(self, item):
        new_text = item.text()
        item_row = self.form.columnList.row(item)

        # Check for duplicate names
        duplicate_found = False
        for i in range(self.form.columnList.count()):
            if i != item_row and self.form.columnList.item(i).text() == new_text:
                duplicate_found = True
                break

        if duplicate_found:
            QtWidgets.QMessageBox.warning(
                self.form,
                translate("Assembly", "Duplicate Name"),
                translate("Assembly", "This name is already used. Please choose a different name."),
            )

            # Revert the change
            old_text = (
                self.bomObj.columnsNames[item_row]
                if self.bomObj and item_row < len(self.bomObj.columnsNames)
                else ""
            )
            item.setText(old_text)
        else:
            isCustomCol = self.isCustomColumn(new_text)

            if not isCustomCol:
                font = item.font()
                font.setBold(True)
                item.setFont(font)
                # Use a single-shot timer to defer changing the flags (else FC crashes)
                QtCore.QTimer.singleShot(0, lambda: self.makeItemNonEditable(item))

            self.updateColumnList()

    def makeItemNonEditable(self, item):
        item.setFlags(item.flags() & ~QtCore.Qt.ItemIsEditable)

    def isNameDuplicate(self, name):
        for i in range(self.form.columnList.count()):
            if self.form.columnList.item(i).text() == name:
                return True
        return False

    def createBomObject(self):
        assembly = UtilsAssembly.activeAssembly()
        Gui.addModule("UtilsAssembly")
        if assembly is not None:
            commands = (
                "assembly = UtilsAssembly.activeAssembly()\n"
                "bom_group = UtilsAssembly.getBomGroup(assembly)\n"
                'bomObj = bom_group.newObject("Assembly::BomObject", "Bill of Materials")'
            )
        else:
            commands = 'bomObj = App.activeDocument().addObject("Assembly::BomObject", "Bill of Materials")'
        Gui.doCommand(commands)
        self.bomObj = Gui.doCommandEval("bomObj")

    def export(self):
        self.bomObj.recompute()
        self.bomObj.ViewObject.exportAsFile()

    def eventFilter(self, watched, event):
        if self.form is not None and watched == self.form.columnList:
            if event.type() == QtCore.QEvent.ShortcutOverride:
                if event.key() == QtCore.Qt.Key_Delete:
                    event.accept()  # Accept the event only if the key is Delete
                    return True  # Indicate that the event has been handled
                return False

            elif event.type() == QtCore.QEvent.KeyPress:
                if event.key() == QtCore.Qt.Key_Delete:
                    selected_indexes = self.form.columnList.selectedIndexes()
                    items_to_remove = []

                    for index in selected_indexes:
                        self.form.columnList.takeItem(index.row())

                    self.updateColumnList()
                    return True  # Consume the event

        return super().eventFilter(watched, event)

    def showHelpDialog(self):
        help_dialog = QtWidgets.QDialog(self.form)
        help_dialog.setWindowFlags(QtCore.Qt.Popup)
        help_dialog.setWindowModality(QtCore.Qt.NonModal)
        help_dialog.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        layout = QtWidgets.QVBoxLayout()
        layout.setContentsMargins(10, 10, 10, 10)

        options_title = QtWidgets.QLabel("<b>" + translate("Assembly", "Options") + "</b>")
        options_text = QtWidgets.QLabel(
            " - "
            + translate(
                "Assembly",
                "Sub-assembly children: the children of sub-assemblies will be included in the bill of materials",
            )
            + "\n"
            " - "
            + translate(
                "Assembly",
                "Parts children: the children of parts will be added to the bill of materials",
            )
            + "\n"
            " - "
            + translate(
                "Assembly",
                "Only parts: adds only part containers and sub-assemblies to the bill of materials. Solids like Part Design bodies, fasteners, or Part workbench primitives are ignored.",
            )
            + "\n"
        )
        columns_title = QtWidgets.QLabel("<b>" + translate("Assembly", "Columns") + "</b>")
        columns_text = QtWidgets.QLabel(
            " - "
            + translate(
                "Assembly",
                "Auto columns :  (Index, Quantity, Name...) are populated automatically. Any modification you make will be overridden. These columns cannot be renamed.",
            )
            + "\n"
            " - "
            + translate(
                "Assembly",
                "Custom columns : 'Description' and other custom columns you add by clicking on 'Add column' will not have their data overwritten. If a column name starts with '.' followed by a property name (e.g. '.Length'), it will be auto-populated with that property value. These columns can be renamed by double-clicking or pressing F2 (renaming a column will currently lose its data).",
            )
            + "\n"
            "\n"
            + translate(
                "Assembly",
                "Any column (custom or not), can be deleted by pressing the Delete key",
            )
            + "\n"
        )
        export_title = QtWidgets.QLabel("<b>" + translate("Assembly", "Export") + "</b>")
        export_text = QtWidgets.QLabel(
            " - "
            + translate(
                "Assembly",
                "The exported file format can be customized in the Spreadsheet workbench preferences",
            )
            + "\n"
        )

        options_text.setWordWrap(True)
        columns_text.setWordWrap(True)
        export_text.setWordWrap(True)

        layout.addWidget(options_title)
        layout.addWidget(options_text)
        layout.addWidget(columns_title)
        layout.addWidget(columns_text)
        layout.addWidget(export_title)
        layout.addWidget(export_text)

        help_dialog.setLayout(layout)
        help_dialog.setFixedWidth(500)

        help_dialog.show()


if App.GuiUp:
    Gui.addCommand("Assembly_CreateBom", CommandCreateBom())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 Victor Titov (DeepSOIC) <vv.titov@gmail.com>       *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This library is free software; you can redistribute it and/or         *
# *   modify it under the terms of the GNU Library General Public           *
# *   License as published by the Free Software Foundation; either          *
# *   version 2 of the License, or (at your option) any later version.      *
# *                                                                         *
# *   This library  is distributed in the hope that it will be useful,      *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this library; see the file COPYING.LIB. If not,    *
# *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
# *   Suite 330, Boston, MA  02111-1307, USA                                *
# *                                                                         *
# ***************************************************************************/

from __future__ import absolute_import

import FreeCAD as App

try:
    from PySide.QtCore import QT_TRANSLATE_NOOP
except ImportError:

    def QT_TRANSLATE_NOOP(ctx, msg):
        return msg


def editAttachment(
    feature=None,
    take_selection=False,
    create_transaction=True,
    callback_OK=None,
    callback_Cancel=None,
    callback_Apply=None,
):
    """Opens attachment editing dialog.
    editAttachment(feature = None,
                   take_selection = False,
                   create_transaction = True,
                   callback_OK = None,
                   callback_Cancel = None,
                   callback_Apply = None)
    feature: object to attach/modify. If None is supplied, the object is taken from
    selection.
    take_selection: if True, current selection is filled into first references (but only
    if object to be attached doesn't have any references already)
    create_transaction = if False, no undo transaction operations will be done by the
    dialog (consequently, canceling the dialog will not reset the feature to original
    state).
    callback_OK: function to be called upon OK. Invoked after writing values to feature,
    committing transaction and closing the dialog.
    callback_Cancel: called after closing the dialog and aborting transaction.
    callback_Apply: invoked after writing values to feature."""

    import AttachmentEditor.TaskAttachmentEditor as TaskAttachmentEditor

    global taskd  # exposing to outside, for ease of debugging
    if feature is None:
        feature = Gui.Selection.getSelectionEx()[0].Object

    try:
        taskd = TaskAttachmentEditor.AttachmentEditorTaskPanel(
            feature,
            take_selection=take_selection,
            create_transaction=create_transaction,
            callback_OK=callback_OK,
            callback_Cancel=callback_Cancel,
            callback_Apply=callback_Apply,
        )
        Gui.Control.showDialog(taskd)
    except TaskAttachmentEditor.CancelError:
        pass


class CommandEditAttachment:
    "Command to edit attachment"

    def GetResources(self):
        return {
            "Pixmap": "Part_Attachment",
            "MenuText": QT_TRANSLATE_NOOP("Part_EditAttachment", "Attachment"),
            "Accel": "",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Part_EditAttachment",
                "Opens the attachment editor to change the attachment of the selected object",
            ),
        }

    def Activated(self):
        try:
            editAttachment()
        except Exception as err:
            from PySide import QtGui

            mb = QtGui.QMessageBox()
            mb.setIcon(mb.Icon.Warning)
            mb.setText(str(err))
            mb.setWindowTitle("Error")
            mb.exec_()

    def IsActive(self):
        sel = Gui.Selection.getSelectionEx()
        if len(sel) == 1:
            if hasattr(sel[0].Object, "Placement"):
                return True
        return False


if App.GuiUp:
    import FreeCADGui as Gui

    command_instance = CommandEditAttachment()
    Gui.addCommand("Part_EditAttachment", command_instance)



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mesh import and export for Z88 mesh file format"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package importZ88Mesh
#  \ingroup FEM
#  \brief FreeCAD Z88 Mesh reader and writer for FEM workbench

import os

import FreeCAD
from FreeCAD import Console

from femmesh import meshtools
from builtins import open as pyopen

# ************************************************************************************************
# ********* generic FreeCAD import and export methods ********************************************
# names are fix given from FreeCAD, these methods are called from FreeCAD
# they are set in FEM modules Init.py


def open(filename):
    """called when freecad opens a file
    a FEM mesh object is created in a new document"""

    docname = os.path.splitext(os.path.basename(filename))[0]
    return insert(filename, docname)


def insert(filename, docname):
    """called when freecad wants to import a file
    a FEM mesh object is created in a existing document"""

    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc

    import_z88_mesh(filename, docname)
    return doc


def export(objectslist, filename):
    "called when freecad exports a file"
    if len(objectslist) != 1:
        Console.PrintError("This exporter can only export one object.\n")
        return
    obj = objectslist[0]
    if not obj.isDerivedFrom("Fem::FemMeshObject"):
        Console.PrintError("No FEM mesh object selected.\n")
        return
    femnodes_mesh = obj.FemMesh.Nodes
    femelement_table = meshtools.get_femelement_table(obj.FemMesh)
    z88_element_type = get_z88_element_type(obj.FemMesh, femelement_table)
    f = pyopen(filename, "w")
    write_z88_mesh_to_file(femnodes_mesh, femelement_table, z88_element_type, f)
    f.close()


# ************************************************************************************************
# ********* module specific methods **************************************************************
# reader:
# - a method uses a FemMesh instance, creates the FreeCAD document object and returns this object
# - a method creates and returns a FemMesh (no FreeCAD document object) out of the FEM mesh dict
# - a method reads the data from the mesh file or converts data and returns FEM mesh dictionary
#
# writer:
# - a method directly writes a FemMesh to the mesh file
# - a method takes a file handle, mesh data and writes to the file handle


# ********* reader *******************************************************************************
def import_z88_mesh(filename, analysis=None, docname=None):
    """read a FEM mesh from a Z88 mesh file and
    insert a FreeCAD FEM Mesh object in the ActiveDocument
    """

    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        try:
            doc = FreeCAD.ActiveDocument
        except NameError:
            doc = FreeCAD.newDocument()
    FreeCAD.ActiveDocument = doc

    mesh_name = os.path.basename(os.path.splitext(filename)[0])

    femmesh = read(filename)
    if femmesh:
        mesh_object = doc.addObject("Fem::FemMeshObject", mesh_name)
        mesh_object.FemMesh = femmesh

    return mesh_object


def read(filename):
    """read a FemMesh from a Z88 mesh file and return the FemMesh"""
    # no document object is created, just the FemMesh is returned

    mesh_data = read_z88_mesh(filename)
    from . import importToolsFem

    return importToolsFem.make_femmesh(mesh_data)


def read_z88_mesh(z88_mesh_input):
    """reads a z88 mesh file z88i1.txt (Z88OSV14) or z88structure.txt (Z88AuroraV3)
    and extracts the nodes and elements
    """
    nodes = {}
    elements_hexa8 = {}
    elements_penta6 = {}
    elements_tetra4 = {}
    elements_tetra10 = {}
    elements_penta15 = {}
    elements_hexa20 = {}
    elements_tria3 = {}
    elements_tria6 = {}
    elements_quad4 = {}
    elements_quad8 = {}
    elements_seg2 = {}
    elements_seg3 = {}

    input_continues = False
    # elem = -1
    z88_element_type = 0

    z88_mesh_file = pyopen(z88_mesh_input, "r")
    mesh_info = z88_mesh_file.readline().strip().split()

    nodes_dimension = int(mesh_info[0])
    nodes_count = int(mesh_info[1])
    elements_count = int(mesh_info[2])
    kflag = int(mesh_info[4])
    # for non rotational elements is --> kflag = 0 --> cartesian, kflag = 1 polar coordinates
    if kflag:
        Console.PrintError("KFLAG = 1, Rotational coordinates not supported at the moment\n")
        return {}
    nodes_first_line = 2  # first line is mesh_info
    nodes_last_line = nodes_count + 1
    elemts_first_line = nodes_last_line + 1
    elements_last_line = elemts_first_line - 1 + elements_count * 2

    Console.PrintLog(f"{nodes_count}\n")
    Console.PrintLog(f"{elements_count}\n")
    Console.PrintLog(f"{nodes_last_line}\n")
    Console.PrintLog(f"{elemts_first_line}\n")
    Console.PrintLog(f"{elements_last_line}\n")

    z88_mesh_file.seek(0)  # go back to the beginning of the file
    for no, line in enumerate(z88_mesh_file):
        lno = no + 1
        linecolumns = line.split()

        if lno >= nodes_first_line and lno <= nodes_last_line:
            # node line
            node_no = int(linecolumns[0])
            node_x = float(linecolumns[2])
            node_y = float(linecolumns[3])
            if nodes_dimension == 2:
                node_z = 0.0
            elif nodes_dimension == 3:
                node_z = float(linecolumns[4])
            nodes[node_no] = FreeCAD.Vector(node_x, node_y, node_z)

        if lno >= elemts_first_line and lno <= elements_last_line:
            # first element line
            if not input_continues:
                elem_no = int(linecolumns[0])
                z88_element_type = int(linecolumns[1])
                input_continues = True

            # second element line
            elif input_continues:
                # not supported elements
                if z88_element_type == 8:
                    # torus8
                    Console.PrintError("Z88 Element No. 8, torus8\n")
                    Console.PrintError("Rotational elements are not supported at the moment\n")
                    return {}
                elif z88_element_type == 12:
                    # torus12
                    Console.PrintError("Z88 Element No. 12, torus12\n")
                    Console.PrintError("Rotational elements are not supported at the moment\n")
                    return {}
                elif z88_element_type == 15:
                    # torus6
                    Console.PrintError("Z88 Element No. 15, torus6\n")
                    Console.PrintError("Rotational elements are not supported at the moment\n")
                    return {}
                elif z88_element_type == 19:
                    # platte16
                    Console.PrintError("Z88 Element No. 19, platte16\n")
                    Console.PrintError("Not supported at the moment\n")
                    return {}
                elif z88_element_type == 21:
                    # schale16, mixture made from hexa8 and hexa20 (thickness is linear)
                    Console.PrintError("Z88 Element No. 21, schale16\n")
                    Console.PrintError("Not supported at the moment\n")
                    return {}
                elif z88_element_type == 22:
                    # schale12, mixtrue made from prism6 and prism15 (thickness is linear)
                    Console.PrintError("Z88 Element No. 22, schale12\n")
                    Console.PrintError("Not supported at the moment\n")
                    return {}

                # supported elements
                elif (
                    z88_element_type == 2
                    or z88_element_type == 4
                    or z88_element_type == 5
                    or z88_element_type == 9
                    or z88_element_type == 13
                    or z88_element_type == 25
                ):
                    # stab4 or stab5 or welle5 or beam13 or beam25 Z88 --> seg2 FreeCAD
                    # N1, N2
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    elements_seg2[elem_no] = (nd1, nd2)
                    input_continues = False
                elif z88_element_type == 3 or z88_element_type == 14 or z88_element_type == 24:
                    # scheibe3 or scheibe14 or schale24 Z88 --> tria6 FreeCAD
                    # N1, N2, N3, N4, N5, N6
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    nd5 = int(linecolumns[4])
                    nd6 = int(linecolumns[5])
                    elements_tria6[elem_no] = (nd1, nd2, nd3, nd4, nd5, nd6)
                    input_continues = False
                elif z88_element_type == 7 or z88_element_type == 20 or z88_element_type == 23:
                    # scheibe7 or platte20 or schale23 Z88 --> quad8 FreeCAD
                    # N1, N2, N3, N4, N5, N6, N7, N8
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    nd5 = int(linecolumns[4])
                    nd6 = int(linecolumns[5])
                    nd7 = int(linecolumns[6])
                    nd8 = int(linecolumns[7])
                    elements_quad8[elem_no] = (nd1, nd2, nd3, nd4, nd5, nd6, nd7, nd8)
                    input_continues = False
                elif z88_element_type == 17:
                    # volume17 Z88 --> tetra4 FreeCAD
                    # N4, N2, N3, N1
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    elements_tetra4[elem_no] = (nd4, nd2, nd3, nd1)
                    input_continues = False
                elif z88_element_type == 16:
                    # volume16 Z88 --> tetra10 FreeCAD
                    # N1, N2, N4, N3, N5, N8, N10, N7, N6, N9
                    # Z88 to FC is different as FC to Z88
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    nd5 = int(linecolumns[4])
                    nd6 = int(linecolumns[5])
                    nd7 = int(linecolumns[6])
                    nd8 = int(linecolumns[7])
                    nd9 = int(linecolumns[8])
                    nd10 = int(linecolumns[9])
                    elements_tetra10[elem_no] = (nd1, nd2, nd4, nd3, nd5, nd8, nd10, nd7, nd6, nd9)
                    input_continues = False
                elif z88_element_type == 1:
                    # volume1 Z88 --> hexa8 FreeCAD
                    # N1, N2, N3, N4, N5, N6, N7, N8
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    nd5 = int(linecolumns[4])
                    nd6 = int(linecolumns[5])
                    nd7 = int(linecolumns[6])
                    nd8 = int(linecolumns[7])
                    elements_hexa8[elem_no] = (nd1, nd2, nd3, nd4, nd5, nd6, nd7, nd8)
                    input_continues = False
                elif z88_element_type == 10:
                    # volume10 Z88 --> hexa20 FreeCAD
                    # N2, N3, N4, N1, N6, N7, N8, N5, N10, N11
                    # N12, N9,  N14, N15, N16, N13, N18, N19, N20, N17
                    # or turn by 90 degree and they match !
                    # N1, N2, N3, N4, N5, N6, N7, N8, N9, N10
                    # N11, N12, N13, N14, N15, N16, N17, N18, N19, N20
                    nd1 = int(linecolumns[0])
                    nd2 = int(linecolumns[1])
                    nd3 = int(linecolumns[2])
                    nd4 = int(linecolumns[3])
                    nd5 = int(linecolumns[4])
                    nd6 = int(linecolumns[5])
                    nd7 = int(linecolumns[6])
                    nd8 = int(linecolumns[7])
                    nd9 = int(linecolumns[8])
                    nd10 = int(linecolumns[9])
                    nd11 = int(linecolumns[10])
                    nd12 = int(linecolumns[11])
                    nd13 = int(linecolumns[12])
                    nd14 = int(linecolumns[13])
                    nd15 = int(linecolumns[14])
                    nd16 = int(linecolumns[15])
                    nd17 = int(linecolumns[16])
                    nd18 = int(linecolumns[17])
                    nd19 = int(linecolumns[18])
                    nd20 = int(linecolumns[19])
                    elements_hexa20[elem_no] = (
                        nd1,
                        nd2,
                        nd3,
                        nd4,
                        nd5,
                        nd6,
                        nd7,
                        nd8,
                        nd9,
                        nd10,
                        nd11,
                        nd12,
                        nd13,
                        nd14,
                        nd15,
                        nd16,
                        nd17,
                        nd18,
                        nd19,
                        nd20,
                    )
                    input_continues = False

                # unknown elements
                # some examples have -1 for some teaching reasons to show some other stuff
                else:
                    Console.PrintError("Unknown element\n")
                    return {}

    for n in nodes:
        Console.PrintLog(str(n) + "  " + str(nodes[n]) + "\n")
    for e in elements_tria6:
        Console.PrintLog(str(e) + "  " + str(elements_tria6[e]) + "\n")
    FreeCAD.Console.PrintLog("\n")

    z88_mesh_file.close()

    return {
        "Nodes": nodes,
        "Seg2Elem": elements_seg2,
        "Seg3Elem": elements_seg3,
        "Tria3Elem": elements_tria3,
        "Tria6Elem": elements_tria6,
        "Quad4Elem": elements_quad4,
        "Quad8Elem": elements_quad8,
        "Tetra4Elem": elements_tetra4,
        "Tetra10Elem": elements_tetra10,
        "Hexa8Elem": elements_hexa8,
        "Hexa20Elem": elements_hexa20,
        "Penta6Elem": elements_penta6,
        "Penta15Elem": elements_penta15,
    }


# ********* writer *******************************************************************************
def write(fem_mesh, filename):
    """directly write a FemMesh to a Z88 mesh file format
    fem_mesh: a FemMesh"""

    if not fem_mesh.isDerivedFrom("Fem::FemMesh"):
        Console.PrintError("Not a FemMesh was given as parameter.\n")
        return
    femnodes_mesh = fem_mesh.Nodes
    femelement_table = meshtools.get_femelement_table(fem_mesh)
    z88_element_type = get_z88_element_type(fem_mesh, femelement_table)
    f = pyopen(filename, "w")
    write_z88_mesh_to_file(femnodes_mesh, femelement_table, z88_element_type, f)
    f.close()


def write_z88_mesh_to_file(femnodes_mesh, femelement_table, z88_element_type, f):
    node_dimension = 3  # 2 for 2D not supported
    if (
        z88_element_type == 4
        or z88_element_type == 17
        or z88_element_type == 16
        or z88_element_type == 1
        or z88_element_type == 10
    ):
        node_dof = 3
    elif z88_element_type == 23 or z88_element_type == 24:
        node_dof = 6  # schalenelemente
    else:
        Console.PrintError("Error: wrong z88_element_type.\n")
        return
    node_count = len(femnodes_mesh)
    element_count = len(femelement_table)
    dofs = node_dof * node_count
    unknown_flag = 0
    written_by = "written by FreeCAD"

    # first line, some z88 specific stuff
    f.write(
        "{} {} {} {} {} {}\n".format(
            node_dimension, node_count, element_count, dofs, unknown_flag, written_by
        )
    )
    # nodes
    for node in femnodes_mesh:
        vec = femnodes_mesh[node]
        f.write(f"{node} {node_dof} {vec.x:.6f} {vec.y:.6f} {vec.z:.6f}\n")
    # elements
    for element in femelement_table:
        # z88_element_type is checked for every element
        # but mixed elements are not supported up to date
        n = femelement_table[element]
        if (
            z88_element_type == 2
            or z88_element_type == 4
            or z88_element_type == 5
            or z88_element_type == 9
            or z88_element_type == 13
            or z88_element_type == 25
        ):
            # seg2 FreeCAD --> stab4 Z88
            # N1, N2
            f.write(f"{element} {z88_element_type}\n")
            f.write(f"{n[0]} {n[1]}\n")
        elif z88_element_type == 3 or z88_element_type == 14 or z88_element_type == 24:
            # tria6 FreeCAD --> schale24 Z88
            # N1, N2, N3, N4, N5, N6
            f.write(f"{element} {z88_element_type}\n")
            f.write(f"{n[0]} {n[1]} {n[2]} {n[3]} {n[4]} {n[5]}\n")
        elif z88_element_type == 7 or z88_element_type == 20 or z88_element_type == 23:
            # quad8 FreeCAD --> schale23 Z88
            # N1, N2, N3, N4, N5, N6, N7, N8
            f.write(f"{element} {z88_element_type}\n")
            f.write(f"{n[0]} {n[1]} {n[2]} {n[3]} {n[4]} {n[5]} {n[6]} {n[7]}\n")
        elif z88_element_type == 17:
            # tetra4 FreeCAD --> volume17 Z88
            # N4, N2, N3, N1
            f.write(f"{element} {z88_element_type}\n")
            f.write(f"{n[3]} {n[1]} {n[2]} {n[0]}\n")
        elif z88_element_type == 16:
            # tetra10 FreeCAD --> volume16 Z88
            # N1, N2, N4, N3, N5, N9, N8, N6, N10, N7, FC to Z88 is different as Z88 to FC
            f.write(f"{element} {z88_element_type}\n")
            f.write(
                "{} {} {} {} {} {} {} {} {} {}\n".format(
                    n[0], n[1], n[3], n[2], n[4], n[8], n[7], n[5], n[9], n[6]
                )
            )
        elif z88_element_type == 1:
            # hexa8 FreeCAD --> volume1 Z88
            # N1, N2, N3, N4, N5, N6, N7, N8
            f.write(f"{element} {z88_element_type}\n")
            f.write(f"{n[0]} {n[1]} {n[2]} {n[3]} {n[4]} {n[5]} {n[6]} {n[7]}\n")
        elif z88_element_type == 10:
            # hexa20 FreeCAD --> volume10 Z88
            # N2, N3, N4, N1, N6, N7, N8, N5, N10, N11
            # N12, N9,  N14, N15, N16, N13, N18, N19, N20, N17
            # or turn by 90 degree and they match !
            # N1, N2, N3, N4, N5, N6, N7, N8, N9, N10
            # N11, N12, N13, N14, N15, N16, N17, N18, N19, N20
            f.write(f"{element} {z88_element_type}\n")
            f.write(
                "{} {} {} {} {} {} {} {} {} {} "
                "{} {} {} {} {} {} {} {} {} {}\n".format(
                    n[0],
                    n[1],
                    n[2],
                    n[3],
                    n[4],
                    n[5],
                    n[6],
                    n[7],
                    n[8],
                    n[9],
                    n[10],
                    n[11],
                    n[12],
                    n[13],
                    n[14],
                    n[15],
                    n[16],
                    n[17],
                    n[18],
                    n[19],
                )
            )
        else:
            Console.PrintError(f"Writing of Z88 elementtype {z88_element_type} not supported.\n")
            # TODO support schale12 (made from prism15) and schale16 (made from hexa20)
            return


# Helper
def get_z88_element_type(femmesh, femelement_table=None):
    return z88_ele_types[meshtools.get_femmesh_eletype(femmesh, femelement_table)]


z88_ele_types = {
    "tetra4": 17,
    "tetra10": 16,
    "hexa8": 1,
    "hexa20": 10,
    "tria3": 0,  # no tria3 available in Z88
    "tria6": 24,
    "quad4": 0,  # no quad4 available in Z88
    "quad8": 23,
    "seg2": 4,  # 3D Truss element
    "seg3": 4,  # 3D Truss element
    "None": 0,
}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Used to create material stock around a machined part - for visualization"""

import FreeCAD
import Path
import math
from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide import QtCore

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class StockType:

    NoStock = "None"
    FromBase = "FromBase"
    CreateBox = "CreateBox"
    CreateCylinder = "CreateCylinder"
    Unknown = "Unknown"

    @classmethod
    def FromStock(cls, stock):
        """FromStock(stock) ... Answer a string representing the type of stock."""
        if not stock:
            return cls.NoStock
        if hasattr(stock, "StockType"):
            return stock.StockType

        # fallback in case somebody messed with internals
        if hasattr(stock, "ExtXneg") and hasattr(stock, "ExtZpos"):
            return cls.FromBase
        if hasattr(stock, "Length") and hasattr(stock, "Width"):
            return cls.CreateBox
        if hasattr(stock, "Radius") and hasattr(stock, "Height"):
            return cls.CreateCylinder
        return cls.Unknown


def shapeBoundBox(obj):
    Path.Log.track(type(obj))
    if isinstance(obj, list) and obj:
        bb = FreeCAD.BoundBox()
        for o in obj:
            bb.add(shapeBoundBox(o))
        return bb

    if hasattr(obj, "Shape"):
        return obj.Shape.BoundBox
    if obj and "App::Part" == obj.TypeId:
        bounds = [shapeBoundBox(o) for o in obj.Group]
        if bounds:
            bb = bounds[0]
            for b in bounds[1:]:
                bb = bb.united(b)
            return bb
    if obj:
        Path.Log.error(translate("PathStock", "Invalid base object %s - no shape found") % obj.Name)
    return None


class Stock(object):
    def onDocumentRestored(self, obj):
        if hasattr(obj, "StockType"):
            obj.setEditorMode("StockType", 2)  # hide

        if hasattr(obj, "Material"):
            obj.removeProperty("Material")
            FreeCAD.Console.PrintWarning(
                translate(
                    "PathStock",
                    "Stock Material property is deprecated. Removing the Material property. Please use native material system to assign a ShapeMaterial",
                )
            )


class StockFromBase(Stock):
    MinExtent = 1

    def __init__(self, obj, base):
        "Make stock"
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "The base object this stock is derived from"),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtXneg",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in negative X-direction",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtXpos",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in positive X-direction",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtYneg",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in negative Y-direction",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtYpos",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in positive Y-direction",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtZneg",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in negative Z-direction",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtZpos",
            "Stock",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra allowance from part bound box in positive Z-direction",
            ),
        )

        obj.Base = base
        obj.ExtXneg = 1.0
        obj.ExtXpos = 1.0
        obj.ExtYneg = 1.0
        obj.ExtYpos = 1.0
        obj.ExtZneg = 1.0
        obj.ExtZpos = 1.0

        # placement is only tracked on creation
        bb = shapeBoundBox(base.Group) if base else None
        if bb:
            obj.Placement = FreeCAD.Placement(
                FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMin), FreeCAD.Rotation()
            )
        else:
            Path.Log.track(obj.Label, base.Label)
        obj.Proxy = self

        # debugging aids
        self.origin = None
        self.length = None
        self.width = None
        self.height = None

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def execute(self, obj):
        bb = shapeBoundBox(obj.Base.Group) if obj.Base and hasattr(obj.Base, "Group") else None
        Path.Log.track(obj.Label, bb)

        # Sometimes, when the Base changes it's temporarily not assigned when
        # Stock.execute is triggered - it'll be set correctly the next time around.
        if bb:
            self.origin = FreeCAD.Vector(-obj.ExtXneg.Value, -obj.ExtYneg.Value, -obj.ExtZneg.Value)

            self.length = bb.XLength + obj.ExtXneg.Value + obj.ExtXpos.Value
            self.width = bb.YLength + obj.ExtYneg.Value + obj.ExtYpos.Value

            if bb.ZLength + obj.ExtZneg.Value + obj.ExtZpos.Value <= 0:
                Path.Log.error("Stock height can not be zero or negative\nSet ExtZneg = 1 mm")
                obj.ExtZneg.Value = self.MinExtent
            self.height = bb.ZLength + obj.ExtZneg.Value + obj.ExtZpos.Value

            shape = Part.makeBox(self.length, self.width, self.height, self.origin)
            shape.Placement = obj.Placement
            obj.Shape = shape

    def onChanged(self, obj, prop):
        if (
            prop in ["ExtXneg", "ExtXpos", "ExtYneg", "ExtYpos", "ExtZneg", "ExtZpos"]
            and "Restore" not in obj.State
        ):
            self.execute(obj)


class StockCreateBox(Stock):
    MinExtent = 0.001

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLength",
            "Length",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Length of this stock box"),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Width",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Width of this stock box"),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Height of this stock box"),
        )

        obj.Length = 10
        obj.Width = 10
        obj.Height = 10

        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def execute(self, obj):
        if obj.Length < self.MinExtent:
            obj.Length = self.MinExtent
        if obj.Width < self.MinExtent:
            obj.Width = self.MinExtent
        if obj.Height < self.MinExtent:
            obj.Height = self.MinExtent

        shape = Part.makeBox(obj.Length, obj.Width, obj.Height)
        shape.Placement = obj.Placement
        obj.Shape = shape

    def onChanged(self, obj, prop):
        if prop in ["Length", "Width", "Height"] and "Restore" not in obj.State:
            self.execute(obj)


class StockCreateCylinder(Stock):
    MinExtent = 0.001

    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLength",
            "Radius",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Radius of this stock cylinder"),
        )
        obj.addProperty(
            "App::PropertyLength",
            "Height",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Height of this stock cylinder"),
        )

        obj.Radius = 2
        obj.Height = 10

        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def execute(self, obj):
        if obj.Radius < self.MinExtent:
            obj.Radius = self.MinExtent
        if obj.Height < self.MinExtent:
            obj.Height = self.MinExtent

        shape = Part.makeCylinder(obj.Radius, obj.Height)
        shape.Placement = obj.Placement
        obj.Shape = shape

    def onChanged(self, obj, prop):
        if prop in ["Radius", "Height"] and "Restore" not in obj.State:
            self.execute(obj)


def SetupStockObject(obj, stockType):
    Path.Log.track(obj.Label, stockType)
    if FreeCAD.GuiUp and obj.ViewObject:
        obj.addProperty(
            "App::PropertyString",
            "StockType",
            "Stock",
            QT_TRANSLATE_NOOP("App::Property", "Internal representation of stock type"),
        )
        obj.StockType = stockType
        obj.setEditorMode("StockType", 2)  # hide

        # If I don't rename the module then usage as Path.Base.Gui.IconViewProvider below
        # 'causes above Path.Log.track(...) to fail with - claiming that Path is accessed
        # before it's assigned.
        # Alternative _another_ `import Path` statement in front of `Path.Log.track(...)`
        # also prevents the issue from happening.
        # Go figure.
        import Path.Base.Gui.IconViewProvider as PathIconViewProvider

        PathIconViewProvider.ViewProvider(obj.ViewObject, "Stock")
        obj.ViewObject.Transparency = 90
        obj.ViewObject.PointSize = 5
        obj.ViewObject.DisplayMode = "Wireframe"


class FakeJob(object):
    def __init__(self, base):
        self.Group = [base]


def _getBase(job):
    if job and hasattr(job, "Model"):
        return job.Model
    if job:
        import PathScripts.PathUtils as PathUtils

        job = PathUtils.findParentJob(job)
        return job.Model if job else None
    return None


def CreateFromBase(job, neg=None, pos=None, placement=None):
    Path.Log.track(job.Label, neg, pos, placement)
    base = _getBase(job)
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Stock")
    obj.Proxy = StockFromBase(obj, base)

    if neg:
        obj.ExtXneg = neg.x
        obj.ExtYneg = neg.y
        obj.ExtZneg = neg.z

    if pos:
        obj.ExtXpos = pos.x
        obj.ExtYpos = pos.y
        obj.ExtZpos = pos.z

    if placement:
        obj.Placement = placement

    SetupStockObject(obj, StockType.FromBase)
    obj.Proxy.execute(obj)
    obj.purgeTouched()
    return obj


def CreateBox(job, extent=None, placement=None):
    base = _getBase(job)
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Stock")
    obj.Proxy = StockCreateBox(obj)

    if extent:
        obj.Length = extent.x
        obj.Width = extent.y
        obj.Height = extent.z
    elif base:
        bb = shapeBoundBox(base.Group)
        obj.Length = max(bb.XLength, 1)
        obj.Width = max(bb.YLength, 1)
        obj.Height = max(bb.ZLength, 1)

    if placement:
        obj.Placement = placement
    elif base:
        bb = shapeBoundBox(base.Group)
        origin = FreeCAD.Vector(bb.XMin, bb.YMin, bb.ZMin)
        obj.Placement = FreeCAD.Placement(origin, FreeCAD.Vector(), 0)

    SetupStockObject(obj, StockType.CreateBox)
    return obj


def CreateCylinder(job, radius=None, height=None, placement=None):
    base = _getBase(job)
    obj = FreeCAD.ActiveDocument.addObject("Part::FeaturePython", "Stock")
    obj.Proxy = StockCreateCylinder(obj)

    if radius:
        obj.Radius = radius

    if height:
        obj.Height = height
    elif base:
        bb = shapeBoundBox(base.Group)
        obj.Radius = math.sqrt(bb.XLength**2 + bb.YLength**2) / 2.0
        obj.Height = max(bb.ZLength, 1)

    if placement:
        obj.Placement = placement
    elif base:
        bb = shapeBoundBox(base.Group)
        origin = FreeCAD.Vector((bb.XMin + bb.XMax) / 2, (bb.YMin + bb.YMax) / 2, bb.ZMin)
        obj.Placement = FreeCAD.Placement(origin, FreeCAD.Vector(), 0)

    SetupStockObject(obj, StockType.CreateCylinder)
    return obj


def TemplateAttributes(stock, includeExtent=True, includePlacement=True):
    attrs = {}
    if stock:
        attrs["version"] = 1
        stockType = StockType.FromStock(stock)
        attrs["create"] = stockType

        if includeExtent:
            if stockType == StockType.FromBase:
                attrs["xneg"] = "%s" % stock.ExtXneg
                attrs["xpos"] = "%s" % stock.ExtXpos
                attrs["yneg"] = "%s" % stock.ExtYneg
                attrs["ypos"] = "%s" % stock.ExtYpos
                attrs["zneg"] = "%s" % stock.ExtZneg
                attrs["zpos"] = "%s" % stock.ExtZpos
            if stockType == StockType.CreateBox:
                attrs["length"] = "%s" % stock.Length
                attrs["width"] = "%s" % stock.Width
                attrs["height"] = "%s" % stock.Height
            if stockType == StockType.CreateCylinder:
                attrs["radius"] = "%s" % stock.Radius
                attrs["height"] = "%s" % stock.Height

        if includePlacement:
            pos = stock.Placement.Base
            attrs["posX"] = pos.x
            attrs["posY"] = pos.y
            attrs["posZ"] = pos.z
            rot = stock.Placement.Rotation
            attrs["rotX"] = rot.Q[0]
            attrs["rotY"] = rot.Q[1]
            attrs["rotZ"] = rot.Q[2]
            attrs["rotW"] = rot.Q[3]

    return attrs


def CreateFromTemplate(job, template):
    if template.get("version") and 1 == int(template["version"]):
        stockType = template.get("create")
        if stockType:
            placement = None
            posX = template.get("posX")
            posY = template.get("posY")
            posZ = template.get("posZ")
            rotX = template.get("rotX")
            rotY = template.get("rotY")
            rotZ = template.get("rotZ")
            rotW = template.get("rotW")
            if (
                posX is not None
                and posY is not None
                and posZ is not None
                and rotX is not None
                and rotY is not None
                and rotZ is not None
                and rotW is not None
            ):
                pos = FreeCAD.Vector(float(posX), float(posY), float(posZ))
                rot = FreeCAD.Rotation(float(rotX), float(rotY), float(rotZ), float(rotW))
                placement = FreeCAD.Placement(pos, rot)
            elif (
                posX is not None
                or posY is not None
                or posZ is not None
                or rotX is not None
                or rotY is not None
                or rotZ is not None
                or rotW is not None
            ):
                Path.Log.warning(
                    "Corrupted or incomplete placement information in template - ignoring"
                )

            if stockType == StockType.FromBase:
                xneg = template.get("xneg")
                xpos = template.get("xpos")
                yneg = template.get("yneg")
                ypos = template.get("ypos")
                zneg = template.get("zneg")
                zpos = template.get("zpos")
                neg = None
                pos = None
                if (
                    xneg is not None
                    and xpos is not None
                    and yneg is not None
                    and ypos is not None
                    and zneg is not None
                    and zpos is not None
                ):
                    neg = FreeCAD.Vector(
                        FreeCAD.Units.Quantity(xneg).Value,
                        FreeCAD.Units.Quantity(yneg).Value,
                        FreeCAD.Units.Quantity(zneg).Value,
                    )
                    pos = FreeCAD.Vector(
                        FreeCAD.Units.Quantity(xpos).Value,
                        FreeCAD.Units.Quantity(ypos).Value,
                        FreeCAD.Units.Quantity(zpos).Value,
                    )
                elif (
                    xneg is not None
                    or xpos is not None
                    or yneg is not None
                    or ypos is not None
                    or zneg is not None
                    or zpos is not None
                ):
                    Path.Log.error(
                        "Corrupted or incomplete specification for creating stock from base - ignoring extent"
                    )
                return CreateFromBase(job, neg, pos, placement)

            if stockType == StockType.CreateBox:
                Path.Log.track(" create box")
                length = template.get("length")
                width = template.get("width")
                height = template.get("height")
                extent = None
                if length is not None and width is not None and height is not None:
                    Path.Log.track("  have extent")
                    extent = FreeCAD.Vector(
                        FreeCAD.Units.Quantity(length).Value,
                        FreeCAD.Units.Quantity(width).Value,
                        FreeCAD.Units.Quantity(height).Value,
                    )
                elif length is not None or width is not None or height is not None:
                    Path.Log.error(
                        "Corrupted or incomplete size for creating a stock box - ignoring size"
                    )
                else:
                    Path.Log.track(
                        "  take placement (%s) and extent (%s) from model" % (placement, extent)
                    )
                return CreateBox(job, extent, placement)

            if stockType == StockType.CreateCylinder:
                radius = template.get("radius")
                height = template.get("height")
                if radius is not None and height is not None:
                    pass
                elif radius is not None or height is not None:
                    radius = None
                    height = None
                    Path.Log.error(
                        "Corrupted or incomplete size for creating a stock cylinder - ignoring size"
                    )
                return CreateCylinder(job, radius, height, placement)

            Path.Log.error(
                translate("PathStock", "Unsupported stock type named {}").format(stockType)
            )
        else:
            Path.Log.error(
                translate("PathStock", "Unsupported PathStock template version {}").format(
                    template.get("version")
                )
            )
        return None


FreeCAD.Console.PrintLog("Loading PathStock... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2022 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Language as PathLanguage
import math

# Path.Log.trackModule(Path.Log.thisModule())

PI = math.pi


class Kink(object):
    """A Kink represents the angle at which two moves connect.
    A positive kink angle represents a move to the left, and a negative angle represents a move to the right.
    """

    def __init__(self, m0, m1):
        if m1 is None:
            m1 = m0[1]
            m0 = m0[0]
        self.m0 = m0
        self.m1 = m1
        self.t0 = m0.anglesOfTangents()[1]
        self.t1 = m1.anglesOfTangents()[0]
        if Path.Geom.isRoughly(self.t0, self.t1):
            self.defl = 0
        else:
            self.defl = Path.Geom.normalizeAngle(self.t1 - self.t0)

    def isKink(self):
        return self.defl != 0

    def goesLeft(self):
        return self.defl > 0

    def goesRight(self):
        return self.defl < 0

    def deflection(self):
        """deflection() ... returns the tangential difference of the two edges at their intersection"""
        return self.defl

    def normAngle(self):
        """normAngle() ... returns the angle opposite between the two tangents"""

        # The normal angle is perpendicular to the "average tangent" of the kink. The question
        # is into which direction to turn. One lies in the center between the two edges and the
        # other is opposite to that. As it turns out, the magnitude of the tangents tell it all.
        if self.t0 > self.t1:
            return Path.Geom.normalizeAngle((self.t0 + self.t1 + math.pi) / 2)
        return Path.Geom.normalizeAngle((self.t0 + self.t1 - math.pi) / 2)

    def position(self):
        """position() ... position of the edge's intersection"""
        return self.m0.positionEnd()

    def x(self):
        return self.position().x

    def y(self):
        return self.position().y

    def __repr__(self):
        return f"({self.x():.4f}, {self.y():.4f})[t0={180*self.t0/math.pi:.2f}, t1={180*self.t1/math.pi:.2f}, deflection={180*self.defl/math.pi:.2f}, normAngle={180*self.normAngle()/math.pi:.2f}]"


class Bone(object):
    """A Bone holds all the information of a bone and the kink it is attached to"""

    def __init__(self, kink, angle, length, instr=None):
        self.kink = kink
        self.angle = angle
        self.length = length
        self.instr = [] if instr is None else instr

    def addInstruction(self, instr):
        self.instr.append(instr)

    def position(self):
        """pos() ... return the position of the bone"""
        return self.kink.position()

    def tip(self):
        """tip() ... return the tip of the bone."""
        dx = abs(self.length) * math.cos(self.angle)
        dy = abs(self.length) * math.sin(self.angle)
        return self.position() + FreeCAD.Vector(dx, dy, 0)


def kink_to_path(kink, g0=False):
    return Path.Path([PathLanguage.instruction_to_command(instr) for instr in [kink.m0, kink.m1]])


def bone_to_path(bone, g0=False):
    kink = bone.kink
    cmds = []
    if g0 and not Path.Geom.pointsCoincide(kink.m0.positionBegin(), FreeCAD.Vector(0, 0, 0)):
        pos = kink.m0.positionBegin()
        param = {}
        if not Path.Geom.isRoughly(pos.x, 0):
            param["X"] = pos.x
        if not Path.Geom.isRoughly(pos.y, 0):
            param["Y"] = pos.y
        cmds.append(Path.Command("G0", param))
    for instr in [kink.m0, bone.instr[0], bone.instr[1], kink.m1]:
        cmds.append(PathLanguage.instruction_to_command(instr))
    return Path.Path(cmds)


def generate_bone(kink, length, angle):
    dx = length * math.cos(angle)
    dy = length * math.sin(angle)
    p0 = kink.position()

    if Path.Geom.isRoughly(0, dx):
        # vertical bone
        moveIn = PathLanguage.MoveStraight(kink.position(), "G1", {"Y": p0.y + dy})
        moveOut = PathLanguage.MoveStraight(moveIn.positionEnd(), "G1", {"Y": p0.y})
    elif Path.Geom.isRoughly(0, dy):
        # horizontal bone
        moveIn = PathLanguage.MoveStraight(kink.position(), "G1", {"X": p0.x + dx})
        moveOut = PathLanguage.MoveStraight(moveIn.positionEnd(), "G1", {"X": p0.x})
    else:
        moveIn = PathLanguage.MoveStraight(kink.position(), "G1", {"X": p0.x + dx, "Y": p0.y + dy})
        moveOut = PathLanguage.MoveStraight(moveIn.positionEnd(), "G1", {"X": p0.x, "Y": p0.y})

    return Bone(kink, angle, length, [moveIn, moveOut])


class Generator(object):
    def __init__(self, calc_length, nominal_length, custom_length):
        self.calc_length = calc_length
        self.nominal_length = nominal_length
        self.custom_length = custom_length

    def length(self, kink, angle):
        return self.calc_length(kink, angle, self.nominal_length, self.custom_length)

    def generate_func(self):
        return generate_bone

    def generate(self, kink):
        angle = self.angle(kink)
        return self.generate_func()(kink, self.length(kink, angle), angle)


class GeneratorTBoneHorizontal(Generator):
    def angle(self, kink):
        if abs(kink.normAngle()) > (PI / 2):
            return -PI
        else:
            return 0


class GeneratorTBoneVertical(Generator):
    def angle(self, kink):
        if kink.normAngle() > 0:
            return PI / 2
        else:
            return -PI / 2


class GeneratorTBoneOnShort(Generator):
    def angle(self, kink):
        rot = PI / 2 if kink.goesRight() else -PI / 2

        if kink.m0.pathLength() < kink.m1.pathLength():
            return Path.Geom.normalizeAngle(kink.t0 + rot)
        else:
            return Path.Geom.normalizeAngle(kink.t1 + rot)


class GeneratorTBoneOnLong(Generator):
    def angle(self, kink):
        rot = PI / 2 if kink.goesRight() else -PI / 2

        if kink.m0.pathLength() > kink.m1.pathLength():
            return Path.Geom.normalizeAngle(kink.t0 + rot)
        else:
            return Path.Geom.normalizeAngle(kink.t1 + rot)


class GeneratorDogbone(Generator):
    def angle(self, kink):
        return kink.normAngle()


def generate(kink, generator, calc_length, nominal_length, custom_length=None):
    if custom_length is None:
        custom_length = nominal_length
    gen = generator(calc_length, nominal_length, custom_length)
    return gen.generate(kink)
